from tests.utils import (
    datagram_from_input,
    standard_datagram_test,
    dg_get_quantity,
    compare_result_dicts,
)


def test_datagram_from_meascsv(datadir):
    input = {
        "case": "data_3.1.0/00-experiment/measurement.csv",
        "parameters": {},
    }
    ret = datagram_from_input(input, "meascsv", datadir)
    standard_datagram_test(ret, {"nsteps": 1, "step": 0, "nrows": 1662, "point": 0})

    val = dg_get_quantity(ret, step="0", col="T_f", utsrow=0)
    compare_result_dicts(val, {"n": 20.9, "s": 0.1, "u": "degC"})

    val = dg_get_quantity(ret, step=0, col="T_f", utsrow=100)
    compare_result_dicts(val, {"n": 455.1, "s": 0.1, "u": "degC"})



"""
**labviewcsv**: Processing Agilent LabVIEW CSV files
----------------------------------------------------

Used to process files generated using Agilent PNA-L N5320C via its LabVIEW driver.
This file format includes a header, with the values of bandwidth and averaging,
and three tab-separated columns containing the frequency :math:`f`, and the real
and imaginary parts of the complex reflection coefficient :math:`\\Gamma(f)`.

Timestamps are determined from file name. One trace per file. As the set-up for
which this format was designed always uses the ``S11`` port, the name of the trace
is hard-coded to this value.

.. codeauthor:: Peter Kraus
"""

from uncertainties.core import str_to_number_with_uncert as tuple_fromstr
import xarray as xr
import datatree


def process(
    *,
    fn: str,
    encoding: str = "utf-8",
    **kwargs: dict,
) -> datatree.DataTree:
    """
    VNA reflection trace parser for Agilent's LabVIEW driver.

    Parameters
    ----------
    fn
        File to process

    encoding
        Encoding of ``fn``, by default "utf-8".

    Returns
    -------
    :class:`datatree.DataTree`
        A :class:`datatree.DataTree` containing a single :class:`xarray.Dataset` with the
        ``S11`` (reflection) trace.

    """

    with open(fn, "r", encoding=encoding) as infile:
        lines = infile.readlines()
    assert (
        len(lines) > 2
    ), f"qftrace: Only {len(lines)-1} points supplied in {fn}; fitting impossible."

    # process header
    bw = [10000.0, 1.0]
    avg = 15
    if ";" in lines[0]:
        items = lines.pop(0).split(";")
        for item in items:
            if item.startswith("BW"):
                bw = tuple_fromstr(item.split("=")[-1].strip())
            if item.startswith("AVG"):
                avg = int(item.split("=")[-1].strip())
    fsbw = bw[0] / avg

    # calculate precision of trace
    freq = {"vals": [], "devs": []}
    real = {"vals": [], "devs": []}
    imag = {"vals": [], "devs": []}
    for line in lines:
        f, re, im = line.strip().split()
        fn, fs = tuple_fromstr(f)
        fs = max(fs, fsbw)
        ren, res = tuple_fromstr(re)
        imn, ims = tuple_fromstr(im)
        freq["vals"].append(fn)
        freq["devs"].append(fs)
        real["vals"].append(ren)
        real["devs"].append(res)
        imag["vals"].append(imn)
        imag["devs"].append(ims)

    vals = xr.Dataset(
        data_vars={
            "Re(G)": (
                ["freq"],
                real["vals"],
                {"ancillary_variables": "Re(G)_std_err"},
            ),
            "Re(G)_std_err": (
                ["freq"],
                real["devs"],
                {"standard_name": "Re(G) standard_error"},
            ),
            "Im(G)": (
                ["freq"],
                imag["vals"],
                {"ancillary_variables": "Im(G)_std_err"},
            ),
            "Im(G)_std_err": (
                ["freq"],
                imag["devs"],
                {"standard_name": "Im(G) standard_error"},
            ),
            "average": avg,
            "bandwidth": (
                [],
                bw[0],
                {"units": "Hz", "ancillary_variables": "bandwidth_std_err"},
            ),
            "bandwidth_std_err": (
                [],
                bw[1],
                {"units": "Hz", "standard_name": "bandwidth standard_error"},
            ),
        },
        coords={
            "freq": (
                ["freq"],
                freq["vals"],
                {"units": "Hz", "ancillary_variables": "freq_std_err"},
            ),
            "freq_std_err": (
                ["freq"],
                freq["devs"],
                {"units": "Hz", "standard_name": "freq standard_error"},
            ),
        },
    )

    return datatree.DataTree.from_dict(dict(S11=vals))



from importlib import metadata
import logging
import importlib
from typing import Callable
from zoneinfo import ZoneInfo
from dgbowl_schemas.yadg.dataschema import DataSchema
from datatree import DataTree
import xarray as xr
import numpy as np
from .. import dgutils, core


datagram_version = metadata.version("yadg")
logger = logging.getLogger(__name__)


def infer_datagram_handler(parser: str) -> tuple[Callable, str]:
    """
    Helper function to distribute work to parsers.

    Parameters
    ----------
    parser
        Name of the parser from schema.

    Returns
    -------
    (process, version): tuple[Callable, str]
        A tuple containing the handler function as :class:`Callable` and the handler
        version as :class:`str`.
    """
    modname = f"yadg.parsers.{parser}"
    try:
        m = importlib.import_module(modname)
        func = getattr(m, "process")
        return func
    except ImportError as e:
        logger.critical(f"could not import module '{modname}'")
        raise e


def process_schema(dataschema: DataSchema, strict_merge: bool = False) -> DataTree:
    """
    Main worker function of **yadg**.

    Takes in a :class:`DataSchema` object and returns a single annotated
    `datagram` created from the `dataschema`. It is the job of the user to supply a
    validated `schema`.

    Parameters
    ----------
    dataschema: DataSchema
        A fully validated `dataschema` object.

    Returns
    -------
    datagram: DataTree
        An unvalidated `datagram`. The `parser`\\ s included in **yadg** should return
        a valid `datagram`; any custom `parser`\\ s might not do so. Use the function
        :meth:`yadg.core.validators.validate_datagram` to validate the resulting `datagram`.
    """

    if strict_merge:
        concatmode = "identical"
    else:
        concatmode = "drop_conflicts"

    root = DataTree()
    root.attrs = {
        "provenance": "yadg process",
        "date": dgutils.now(asstr=True),
        "input_schema": dataschema.json(),
        "datagram_version": core.datagram_version,
    }
    root.attrs.update(dgutils.get_yadg_metadata())

    while hasattr(dataschema, "update"):
        dataschema = dataschema.update()
        if hasattr(dataschema, "metadata"):
            if hasattr(dataschema.metadata, "version"):
                if dataschema.metadata.version == "5.0":
                    break

    for si, step in enumerate(dataschema.steps):
        logger.info("Processing step %d:", si)

        # Backfill default timezone, locale, encoding.
        if step.extractor.timezone is None:
            tz = ZoneInfo(dataschema.step_defaults.timezone)
        else:
            tz = ZoneInfo(step.extractor.timezone)
        if step.extractor.locale is None:
            loc = dataschema.step_defaults.locale
        else:
            loc = step.extractor.locale
        if step.extractor.encoding is None:
            enc = dataschema.step_defaults.encoding
        else:
            enc = step.extractor.encoding

        if step.tag is None:
            step.tag = f"{si}"

        handler = infer_datagram_handler(step.parser)
        todofiles = step.input.paths()
        if len(todofiles) == 0:
            logger.warning("No files processed by step '%s'.", step.tag)
        vals = None
        for tf in todofiles:
            logger.info("Processing file '%s'.", tf)
            fvals = handler(
                fn=tf,
                encoding=enc,
                timezone=tz,
                locale=loc,
                filetype=step.extractor.filetype,
                parameters=step.parameters,
            )
            if isinstance(fvals, DataTree):
                tasks = [(k, v.to_dataset()) for k, v in fvals.items()]
            else:
                tasks = [(None, fvals)]
            for task in tasks:
                name, dset = task
                for k in dset:
                    assert (
                        "/" not in k
                    ), f"The character '/' cannot be used in variable names: '{k}'."
                if not hasattr(dset, "uts"):
                    dset = dset.expand_dims("uts")
                if len(dset.uts.coords) == 0:
                    dset["uts"] = np.zeros(dset.uts.size)
                    dset.attrs["fulldate"] = False
                if (
                    not dset.attrs.get("fulldate", True)
                    or step.externaldate is not None
                ):
                    ts, fulldate = dgutils.complete_timestamps(
                        timesteps=dset.uts.values,
                        fn=tf,
                        spec=step.externaldate,
                        timezone=tz,
                    )
                    dset["uts"] = ts
                    if fulldate:
                        dset.attrs.pop("fulldate", None)
                    else:
                        dset.attrs["fulldate"] = int(fulldate)
                if name is None:
                    fvals = dset
                else:
                    fvals[name] = DataTree(dset)
            if vals is None:
                vals = fvals
            elif isinstance(vals, xr.Dataset):
                try:
                    vals = xr.concat([vals, fvals], dim="uts", combine_attrs=concatmode)
                except xr.MergeError:
                    raise RuntimeError(
                        "Merging metadata from multiple files has failed, as some of the "
                        "values differ between files. This might be caused by trying to "
                        "parse data obtained using different techniques/protocols in a "
                        "single step. If you are certain this is what you want, try using "
                        "yadg with the '--ignore-merge-errors' option."
                    )
            elif isinstance(vals, DataTree):
                for k, v in fvals.items():
                    if k in vals:  # pylint: disable=E1135
                        try:
                            newv = xr.concat(
                                [vals[k].ds, v.ds],  # pylint: disable=E1136
                                dim="uts",
                                combine_attrs=concatmode,
                            )
                        except xr.MergeError:
                            raise RuntimeError(
                                "Merging metadata from multiple files has failed, as some of the "
                                "values differ between files. This might be caused by trying to "
                                "parse data obtained using different techniques/protocols in a "
                                "single step. If you are certain this is what you want, try using "
                                "yadg with the '--ignore-merge-errors' option."
                            )
                    else:
                        newv = v.ds
                    vals[k] = DataTree(newv)  # pylint: disable=E1137
        if isinstance(vals, xr.Dataset):
            stepdt = DataTree.from_dict({"/": vals})
        elif isinstance(vals, DataTree):
            stepdt = vals
        else:
            stepdt = DataTree()
        stepdt.name = step.tag
        stepdt.attrs = dict(parser=step.parser)
        stepdt.parent = root
    return root



"""
This module handles the reading and processing of X-ray diffraction data. It loads X-ray
diffraction data, determines reasonable uncertainties of the signal intensity (y-axis),
and explicitly populates the angle axis (:math:`2\\theta`), if necessary.

Usage
`````
Available since ``yadg-4.0``. The parser supports the following parameters:

.. _yadg.parsers.xrdtrace.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_0.step.XRDTrace

.. _yadg.parsers.xrdtrace.formats:

Formats
```````
The ``filetypes`` currently supported by the parser are:

 - PANalytical ``xrdml`` files (``panalytical.xrdml``),
   see :mod:`~yadg.parsers.xrdtrace.panalyticalxrdml`
 - PANalytical ``csv`` files (``panalytical.csv``),
   see :mod:`~yadg.parsers.xrdtrace.panalyticalcsv`
 - PANalytical ``xy`` files (``panalytical.xy``),
   see :mod:`~yadg.parsers.xrdtrace.panalyticalxy`

.. _yadg.parsers.xrdtrace.provides:

Provides
````````
The raw data is stored, for each timestep, using the following format:

.. code-block:: yaml

  xr.Dataset:
    coords:
      uts:         !!float
      angle:       !!float            # Diffraction angle (deg)
    data_vals:
      intensity:   (uts, angle)       # Detector intensity (counts)

"""

import xarray as xr
from . import panalyticalxrdml, panalyticalcsv, panalyticalxy


def process(
    *,
    filetype: str,
    **kwargs: dict,
) -> xr.Dataset:
    """
    Unified X-ray diffractogram data parser. Forwards ``kwargs`` to the worker
    functions based on the supplied ``filetype``.

    This parser processes XPS scans in signal(energy) format.

    Parameters
    ----------
    filetype
        Discriminator used to select the appropriate worker function.

    Returns
    -------
    :class:`xarray.Dataset`


    """
    if filetype == "panalytical.xrdml":
        return panalyticalxrdml.process(**kwargs)
    elif filetype == "panalytical.csv":
        return panalyticalcsv.process(**kwargs)
    elif filetype == "panalytical.xy":
        return panalyticalxy.process(**kwargs)



from .helpers import get_yadg_metadata
from .dateutils import now, infer_timestamp_from, ole_to_uts, complete_timestamps
from .utils import update_schema, schema_from_preset
from .btools import read_value
from .pintutils import sanitize_units, ureg

__all__ = [
    "get_yadg_metadata",
    "now",
    "infer_timestamp_from",
    "ole_to_uts",
    "complete_timestamps",
    "update_schema",
    "schema_from_preset",
    "read_value",
    "sanitize_units",
    "ureg",
]



import pytest
from tests.utils import datagram_from_input, standard_datagram_test


@pytest.mark.parametrize(
    "input, ts",
    [
        (
            {  # ts1 - BOM requires utf-8-sig encoding
                "case": "log 2021-09-17 11-26-14.140.csv",
                "parser": "basiccsv",
                "encoding": "utf-8-sig",
                "parameters": {
                    "timestamp": {
                        "timestamp": {"index": 0, "format": '"%Y-%m-%d %H:%M:%S.%f"'}
                    },
                    "units": {},
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 239,
                "point": 0,
                "pars": {"uts": {"value": 1631877974.045}},
                "errortype": ValueError,
                "errorexpr": "ufeff",
            },
        ),
    ],
)
def test_datagram(input, ts, datadir):
    ret = datagram_from_input(input, input["parser"], datadir)
    standard_datagram_test(ret, ts)

    fail_input = input.copy()
    del fail_input["encoding"]
    with pytest.raises(ts["errortype"], match=ts["errorexpr"]):
        datagram_from_input(fail_input, fail_input["parser"], datadir)



"""
panalyticalxy: Processing of PANalytical XRD ``xy`` files
---------------------------------------------------------

File Structure
``````````````

These files basically just contain the ``[Scan points]`` part of
:mod:`~yadg.parsers.xrdtrace.panalyticalcsv` files. As a consequence, no metadata
is recorded, and the format does not have an associated timestamp.

Uncertainties
`````````````
The uncertainties of ``"angle"`` are taken from the number of significant figures.

The uncertainties of ``"intensity"`` are taken from the number of significant figures.

.. codeauthor::
    Nicolas Vetsch,
    Peter Kraus
"""

from uncertainties.core import str_to_number_with_uncert as tuple_fromstr
import numpy as np
import xarray as xr


def process(
    *,
    fn: str,
    encoding: str,
    **kwargs: dict,
) -> xr.Dataset:
    """Processes a PANalytical XRD xy file.

    Parameters
    ----------
    fn
        The file containing the trace(s) to parse.

    encoding
        Encoding of ``fn``, by default "utf-8".


    Returns
    -------
    :class:`xarray.Dataset`
        Tuple containing the timesteps and metadata. A full timestamp is not available
        in ``.xy`` files.

    """
    with open(fn, "r", encoding=encoding) as xy_file:
        xy = xy_file.readlines()
    datapoints = [li.strip().split() for li in xy]
    angle, intensity = list(zip(*datapoints))
    angle, _ = list(zip(*[tuple_fromstr(a) for a in angle]))
    insty, _ = list(zip(*[tuple_fromstr(i) for i in intensity]))
    idevs = np.ones(len(insty))
    adiff = np.abs(np.diff(angle)) * 0.5
    adiff = np.append(adiff, adiff[-1])
    vals = xr.Dataset(
        data_vars={
            "intensity": (
                ["angle"],
                list(insty),
                {"units": "counts", "ancillary_variables": "intensity_std_err"},
            ),
            "intensity_std_err": (
                ["angle"],
                idevs,
                {"units": "counts", "standard_name": "intensity standard_error"},
            ),
            "angle_std_err": (
                ["angle"],
                adiff,
                {"units": "deg", "standard_name": "angle standard_error"},
            ),
        },
        coords={
            "angle": (
                ["angle"],
                list(angle),
                {"units": "deg", "ancillary_variables": "angle_std_err"},
            ),
        },
    )
    return vals



"""
This is a dummy parser, used mainly for testing of the :mod:`yadg` and :mod:`tomato`
packages. It provides no real functionality.

Usage
`````
Available since ``yadg-4.0``. The parser supports the following parameters:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_0.step.Dummy

Formats
```````
The ``filetypes`` currently supported by the parser are:

 - tomato's JSON file (``tomato.json``)

Schema
``````
The output schema is only defined for the ``tomato.json`` filetype.

.. code-block:: yaml

  xr.Dataset:
    coords:
      uts:           !!float
    data_vars:
      {{ entries }}  (uts)        # Elements present in the "data" entry

The value of every element of ``data`` is assigned a deviation of 0.0.

Module Functions
````````````````

"""

from pydantic import BaseModel
import json
from ... import dgutils
from ..basiccsv.main import append_dicts, dicts_to_dataset
from datatree import DataTree


def process(
    *,
    fn: str,
    filetype: str,
    parameters: BaseModel,
    **kwargs: dict,
) -> DataTree:
    """
    A dummy parser.

    This parser simply returns the current time, the filename provided, and any
    ``kwargs`` passed.

    In case the provided ``filetype`` is a ``tomato.json`` file, this is a json
    data file from the :mod:`tomato` package, which should contain a :class:`list`
    of ``{"value": float, "time": float}`` datapoints in its ``data`` entry.

    Parameters
    ----------
    fn
        Filename to process

    filetype
        Accepts ``tomato.json`` as an optional "dummy instrument" filetype from
        :mod:`tomato`.

    parameters
        Parameters for :class:`~dgbowl_schemas.yadg.dataschema_5_0.step.Dummy`.

    Returns
    -------
    :class:`xarray.Dataset`

    """
    if filetype == "tomato.json":
        with open(fn, "r") as inf:
            jsdata = json.load(inf)

        data_vals = {}
        meta_vals = {}
        for vi, vals in enumerate(jsdata["data"]):
            vals["uts"] = vals.pop("time")
            devs = {}
            for k, v in vals.items():
                if k not in {"time", "address", "channel"}:
                    devs[k] = 0.0
            append_dicts(vals, devs, data_vals, meta_vals, fn, vi)
    else:
        kwargs = {} if parameters is None else parameters.dict()
        if "parser" in kwargs:
            del kwargs["parser"]
        data_vals = {k: [v] for k, v in kwargs.items()}
        data_vals["uts"] = [dgutils.now()]
        meta_vals = {}

    return dicts_to_dataset(data_vals, meta_vals, fulldate=False)



"""
**fusionzip**: Processing Inficon Fusion zipped data format (zip).
------------------------------------------------------------------

This is a wrapper parser which unzips the provided zip file, and then uses
the :mod:`yadg.parsers.chromdata.fusionjson` parser to parse every data
file present in the archive.

.. codeauthor:: Peter Kraus
"""

import zipfile
import tempfile
import os
import xarray as xr

from .fusionjson import process as processjson


def process(*, fn: str, encoding: str, timezone: str, **kwargs: dict) -> xr.Dataset:
    """
    Fusion zip file format.

    The Fusion GC's can export their json formats as a zip archive of a folder
    of jsons. This parser allows for parsing of this zip archive directly,
    without the user having to unzip & move the data.

    Parameters
    ----------
    fn
        Filename to process.

    encoding
        Not used as the file is binary.

    timezone
        Timezone information. This should be ``"localtime"``.

    Returns
    -------
    :class:`xarray.Dataset`
        The data from the inidividual json files contained in the zip archive are
        concatenated into a single :class:`xarray.Dataset`. This might fail if the metadata
        in the json files differs, or if the dimensions are not easily concatenable.

    """

    zf = zipfile.ZipFile(fn)
    with tempfile.TemporaryDirectory() as tempdir:
        zf.extractall(tempdir)
        ds = None
        for ffn in sorted(os.listdir(tempdir)):
            ffn = os.path.join(tempdir, ffn)
            if ffn.endswith("fusion-data"):
                ids = processjson(fn=ffn, encoding=encoding, timezone=timezone)
                if ds is None:
                    ds = ids
                else:
                    try:
                        ds = xr.concat([ds, ids], dim="uts", combine_attrs="identical")
                    except xr.MergeError:
                        raise RuntimeError(
                            "Merging metadata from the unzipped fusion-json files has failed. "
                            "This might be caused by trying to parse data obtained using "
                            "different chromatographic methods. Please check the contents "
                            "of the unzipped files."
                        )
    return ds



import pytest
import os
import json
import datatree
import yadg.core
from dgbowl_schemas.yadg import to_dataschema, DataSchema_4_0, DataSchema_4_1
from pydantic import ValidationError
from pydantic.v1 import ValidationError as ValidationError_v1

from .schemas import ts0, ts1, ts2, ts3, ts4, ts5
from .schemas import fts0, fts1, fts2, fts3, fts4, fts5, fts6, fts7, fts8

from .utils import datagram_from_file, standard_datagram_test, pars_datagram_test


@pytest.mark.parametrize(
    "inp_dict, l_dg, l_res",
    [
        (ts0, 1, 0),
        (ts1, 1, 1),
        (ts2, 1, 2),
        (ts3, 1, 2),
        (ts4, 1, 1),
        (ts5, 1, 0),
    ],
)
def test_datagram_from_schema_dict(inp_dict, l_dg, l_res, datadir):
    os.chdir(datadir)
    ds = to_dataschema(**inp_dict)
    ret = yadg.core.process_schema(ds)
    print(ret)
    assert len(ret.children) == l_dg, "wrong number of steps"
    if l_dg > 0:
        assert len(ret["0"].get("uts", [])) == l_res, "wrong number of timesteps"
    ret.to_netcdf("test.nc", engine="h5netcdf")
    ref = datatree.open_datatree("test.nc")
    print(ref)
    assert ref == ret


@pytest.mark.parametrize(
    "inp_fn, ts",
    [
        (
            "ts0.dummy.json",
            {"nsteps": 1, "step": "0", "item": 0, "kwargs": {}},
        ),
        (
            "ts1.dummy.json",
            {"nsteps": 2, "step": "1", "item": 0, "kwargs": {"k": "v"}},
        ),
        (
            "ts2.json",
            {"nsteps": 2, "step": "1", "item": 0, "kwargs": {"k": "v"}},
        ),
    ],
)
def test_datagram_from_schema_file(inp_fn, ts, datadir):
    os.chdir(datadir)
    jsonpath = datadir.join(inp_fn)
    with open(jsonpath, "r") as infile:
        schema = json.load(infile)
    ds = to_dataschema(**schema)
    ret = yadg.core.process_schema(ds)
    assert len(ret.children) == ts["nsteps"], (
        "wrong number of steps: " f"got: {len(ret.children)}, expected: {ts['nsteps']}"
    )
    for k, v in ts["kwargs"].items():
        assert ret[ts["step"]][k][ts["item"]] == v, "kwargs not passed correctly"
    ret.to_netcdf("test.nc", engine="h5netcdf")
    ref = datatree.open_datatree("test.nc")
    print(ref)
    assert ref == ret


@pytest.mark.parametrize(
    "inp_dict, expr",
    [
        (fts0, r"Discriminator 'parser' is missing in value"),
        (fts1, r"No match for discriminator 'parser' and value 'dumm'"),
        (fts2, r"Both 'files' and 'folders'"),
        (fts3, r"Neither 'files' nor 'folders'"),
        (fts4, r"steps -> 0 -> Dummy -> key\n  extra fields not permitted"),
    ],
)
def test_schema_validator_4_0(inp_dict, expr, datadir):
    os.chdir(datadir)
    with pytest.raises((ValidationError, ValidationError_v1), match=expr):
        assert DataSchema_4_0(**inp_dict)


@pytest.mark.parametrize(
    "inp_dict, expr",
    [
        (fts5, r"metadata -> provenance\n  value is not a valid dict"),
        (fts6, r"metadata -> version\n  unexpected value"),
        (fts7, r"steps -> 0 -> Dummy -> input\n  field required"),
        (fts8, r"Discriminator 'parser' is missing in value"),
    ],
)
def test_schema_validator_4_1(inp_dict, expr, datadir):
    os.chdir(datadir)
    with pytest.raises((ValidationError, ValidationError_v1), match=expr):
        assert DataSchema_4_1(**inp_dict)


@pytest.mark.parametrize(
    "infile, ts",
    [
        (
            "ts0_dummy_tomato.yml",
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 20,
                "point": 9,
                "pars": {
                    "uts": {"value": 100000000.905299},
                    "value": {"value": 0.485624320335729, "sigma": 0.0, "unit": None},
                },
            },
        ),
    ],
)
def test_dummy_tomato(infile, ts, datadir):
    ret = datagram_from_file(infile, datadir)
    standard_datagram_test(ret, ts)
    pars_datagram_test(ret, ts)



import pytest
from tests.utils import (
    datagram_from_input,
    standard_datagram_test,
    pars_datagram_test,
)


@pytest.mark.parametrize(
    "input, ts",
    [
        (
            {  # ts0 - plain time, float offset
                "case": "case_time_custom.csv",
                "parameters": {
                    "timestamp": {"time": {"index": 0, "format": "%I.%M%p"}}
                },
                "externaldate": {"from": {"utsoffset": 0.0}},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 3,
                "point": 0,
                "pars": {
                    "flow": {"sigma": 0.1, "value": 15.0, "unit": "ml/min"},
                    "uts": {"value": 43140.0},
                },
            },
        ),
        (
            {  # ts1 - plain time, string offset
                "case": "case_time_custom.csv",
                "parameters": {
                    "timestamp": {"time": {"index": 0, "format": "%I.%M%p"}}
                },
                "externaldate": {"from": {"isostring": "2021-01-01"}},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 3,
                "point": 0,
                "pars": {"uts": {"value": 1609502340.0}},
            },
        ),
        (
            {  # ts2 - plain time, json file, iso string, implied partial
                "case": "case_time_custom.csv",
                "parameters": {
                    "timestamp": {"time": {"index": 0, "format": "%I.%M%p"}}
                },
                "externaldate": {
                    "from": {"file": {"path": "file_2.json", "type": "json"}}
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 3,
                "point": 0,
                "pars": {"uts": {"value": 1262347140.0}},
            },
        ),
        (
            {  # ts3 - plain time, json file, iso string, full replace
                "case": "case_time_custom.csv",
                "parameters": {
                    "timestamp": {"time": {"index": 0, "format": "%I.%M%p"}}
                },
                "externaldate": {
                    "from": {"file": {"path": "file_3.json", "type": "json"}},
                    "mode": "replace",
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 3,
                "point": 0,
                "pars": {"uts": {"value": 1262433540.0}},
            },
        ),
        (
            {  # ts4 - plain time, pickle file, string, full replace
                "case": "case_time_custom.csv",
                "parameters": {
                    "timestamp": {"time": {"index": 0, "format": "%I.%M%p"}}
                },
                "externaldate": {
                    "from": {"file": {"path": "file_4.pkl", "type": "pkl"}},
                    "mode": "replace",
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 3,
                "point": 0,
                "pars": {"uts": {"value": 1641810718.0}},
            },
        ),
        (
            {  # ts5 - plain time, pickle file, np.ones(3), partial
                "case": "case_time_custom.csv",
                "parameters": {
                    "timestamp": {"time": {"index": 0, "format": "%I.%M%p"}}
                },
                "externaldate": {
                    "from": {"file": {"path": "file_5.pkl", "type": "pkl"}},
                    "mode": "add",
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 3,
                "point": 0,
                "pars": {"uts": {"value": 43141.0}},
            },
        ),
        (
            {  # ts6 -
                "case": "case_time_custom.csv",
                "parameters": {
                    "timestamp": {"time": {"index": 0, "format": "%I.%M%p"}}
                },
                "externaldate": {
                    "from": {
                        "file": {"path": "file_6.json", "type": "json", "match": "uts"}
                    },
                    "mode": "add",
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 3,
                "point": 0,
                "pars": {"uts": {"value": 10000043140.0}},
            },
        ),
    ],
)
def test_datagram(input, ts, datadir):
    ret = datagram_from_input(input, "basiccsv", datadir)
    standard_datagram_test(ret, ts)
    pars_datagram_test(ret, ts)



import json
import os
import pytest

from tests.utils import (
    datagram_from_input,
    standard_datagram_test,
    dg_get_quantity,
    compare_result_dicts,
)

# Tests for the quadstar parser:
# TODO: Implement more thorough tests.


@pytest.mark.parametrize(
    "input, ts",
    [
        (  # ts0 - airdemo.sac
            {"case": "airdemo.sac", "parameters": {"tracetype": "quadstar.sac"}},
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 21,
                "point": 0,
            },
        ),
        (  # ts1 - test.sac
            {"case": "test.sac", "parameters": {"tracetype": "quadstar.sac"}},
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 188,
                "point": 0,
            },
        ),
    ],
)
def test_datagram_from_quadstar(input, ts, datadir):
    ret = datagram_from_input(input, "masstrace", datadir)
    standard_datagram_test(ret, ts)


@pytest.mark.parametrize(
    "input",
    [
        (
            {  # ts0 - test.sac
                "case": "test.sac",
                "parameters": {"tracetype": "quadstar.sac"},
            }
        ),
    ],
)
def test_masstrace_compare_raw_values(input, datadir):
    os.chdir(datadir)
    with open("test_traces.json", "r") as infile:
        ref = json.load(infile)["traces"]
    dg = datagram_from_input(input, "masstrace", datadir)
    step = dg["0"]
    for k in ["1", "2"]:
        for kk in ["mass_to_charge", "y"]:
            print(f"{k=}, {kk=}")
            ret = dg_get_quantity(step, k, col=kk, utsrow=0)
            compare_result_dicts(ret, ref[k][kk])



"""
This module handles the reading and processing of X-ray photoelectron spectroscopy
data, including determining the uncertainties of the signal (y-axis), and explicitly
populating the points in the energy axis (``E``).

Usage
`````
Available since ``yadg-4.1``. The parser supports the following parameters:

.. _yadg.parsers.xpstrace.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_0.step.XPSTrace

.. _yadg.parsers.xpstrace.formats:

Formats
```````
The ``filetypes`` currently supported by the parser are:

 - ULVAC PHI Multipak XPS traces (``phi.spe``),
   see :mod:`~yadg.parsers.xpstrace.phispe`

.. _yadg.parsers.xpstrace.provides:

Provides
````````
The raw data is stored, for each timestep, using the following format:

.. code-block:: yaml

  datatree.DataTree:
    {{ trace_name }}      !!xr.Dataset
      coords:
        uts:              !!float
        E:                !!float      # binding energies (eV)
      data_vals:
        y:                (uts, E)     # signal

Module Functions
````````````````

"""

import datatree
from . import phispe


def process(
    *,
    filetype: str,
    **kwargs: dict,
) -> datatree.DataTree:
    """
    Unified x-ray photoelectron spectroscopy parser. Forwards ``kwargs`` to the worker
    functions based on the supplied ``filetype``.

    This parser processes XPS scans in signal(energy) format.

    Parameters
    ----------
    filetype
        Discriminator used to select the appropriate worker function.

    Returns
    -------
    :class:`datatree.DataTree`

    """
    if filetype == "phi.spe":
        return phispe.process(**kwargs)



import json
import os
import pytest

from tests.utils import (
    datagram_from_input,
    standard_datagram_test,
    compare_result_dicts,
    dg_get_quantity,
)


@pytest.mark.parametrize(
    "input, ts",
    [
        (  # ts0 - test0.spe
            {"case": "test0.spe", "parameters": {"tracetype": "phi.spe"}},
            {"nsteps": 1, "step": 0, "nrows": 1, "point": 0},
        ),
        (  # ts1 - test1.spe
            {"case": "test1.spe", "parameters": {"tracetype": "phi.spe"}},
            {"nsteps": 1, "step": 0, "nrows": 1, "point": 0},
        ),
        (  # ts2 - check concatenation
            {
                "files": ["test0.spe", "test0b.spe"],
                "parameters": {"tracetype": "phi.spe"},
            },
            {"nsteps": 1, "step": 0, "nrows": 2, "point": 0},
        ),
    ],
)
def test_datagram_from_xpstrace(input, ts, datadir):
    ret = datagram_from_input(input, "xpstrace", datadir)
    print(f"{ret=}")
    standard_datagram_test(ret, ts)


@pytest.mark.parametrize(
    "input",
    [
        (
            {  # ts0 - test0.spe
                "case": "test0.spe",
                "parameters": {"tracetype": "phi.spe"},
            }
        ),
    ],
)
def test_xpstrace_compare_raw_values(input, datadir):
    os.chdir(datadir)
    with open("test0.json", "r") as infile:
        ref = json.load(infile)["traces"]
    dg = datagram_from_input(input, "xpstrace", datadir)
    step = dg["0"]
    for k in {"1su", "F1s"}:
        for kk in {"y", "E"}:
            ret = dg_get_quantity(step, k, col=kk, utsrow=0)
            compare_result_dicts(ret, ref[k][kk])



"""
Handles the reading and processing of flow controller or flow meter data.

Usage
`````
Available since ``yadg-4.0``. The parser supports the following parameters:

.. _yadg.parsers.flowdata.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_0.step.FlowData

.. _yadg.parsers.flowdata.formats:

Formats
```````
The ``filetypes`` currently supported by the parser are:

  - DryCal log file text output (``drycal.txt``),
    see :mod:`~yadg.parsers.flowdata.drycal`
  - DryCal log file tabulated output (``drycal.csv``),
    see :mod:`~yadg.parsers.flowdata.drycal`
  - DryCal log file document file (``drycal.rtf``),
    see :mod:`~yadg.parsers.flowdata.drycal`

.. _yadg.parsers.flowdata.provides:

Schema
``````
The parser is used to extract all tabular data in the input file. This parser processes
additional calibration information analogously to :mod:`~yadg.parsers.basiccsv`.

Module Functions
````````````````

"""

from .main import process

__all__ = ["process"]



import pytest
import os
import json
from tests.utils import (
    datagram_from_input,
    standard_datagram_test,
    compare_result_dicts,
    dg_get_quantity,
)


@pytest.mark.parametrize(
    "input, ts",
    [
        (
            {  # ts0 - 01-inert
                "folders": ["data_3.1.0/01-inert"],
                "parameters": {},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 1,
                "point": 0,
                "tracelen": 20001,
                "npeaks": 2,
                "peak": 0,
                "test": {
                    "S11": {
                        "i": 0,
                        "f": {"n": 7.1e9, "s": 1e3, "u": "Hz"},
                        "re": {"n": -0.0192804, "s": 1e-8, "u": " "},
                        "im": {"n": 0.9448405, "s": 1e-7, "u": " "},
                    }
                },
            },
        )
    ],
)
def test_datagram_from_qftrace(input, ts, datadir):
    os.chdir(datadir)
    dg = datagram_from_input(input, "qftrace", datadir)
    step = dg["0"]
    standard_datagram_test(step, ts)
    for k, v in {
        "freq": {"n": 7.1e9, "s": 1e3, "u": "Hz"},
        "Re(G)": {"n": -0.0192804, "s": 1e-8, "u": None},
        "Im(G)": {"n": 0.9448405, "s": 1e-7, "u": None},
    }.items():
        ret = dg_get_quantity(step, "S11", col=k, utsrow=0)
        print(f"{ret=}")
        point = {"n": ret["n"][0], "s": ret["s"][0], "u": ret["u"]}
        compare_result_dicts(point, v)


def test_qftrace_compare_raw_values(datadir):
    input = {
        "folders": ["data_3.1.0/01-inert"],
        "parameters": {},
    }
    os.chdir(datadir)
    dg = datagram_from_input(input, "qftrace", datadir)
    step = dg["0"]
    with open("yvals.json", "r") as infile:
        ref = json.load(infile)["traces"]["S11"]
    for k in {"freq", "Re(G)", "Im(G)"}:
        ret = dg_get_quantity(step, "S11", col=k, utsrow=0)
        print(f"{ret=}")
        compare_result_dicts(ret, ref[k])



import logging
import xarray as xr
from zoneinfo import ZoneInfo
from . import drycal

logger = logging.getLogger(__name__)


def process(
    *,
    fn: str,
    filetype: str,
    encoding: str,
    timezone: ZoneInfo,
    **kwargs: dict,
) -> xr.Dataset:
    """
    Flow meter data processor

    This parser processes flow meter data.

    Parameters
    ----------
    fn
        File to process

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "localtime".

    parameters
        Parameters for :class:`~dgbowl_schemas.yadg.dataschema_5_0.step.FlowData`.

    Returns
    -------
    :class:`xarray.Dataset`

    """

    if filetype.startswith("drycal"):
        if filetype.endswith(".rtf") or fn.endswith("rtf"):
            vals = drycal.rtf(fn, encoding, timezone)
        elif filetype.endswith(".csv") or fn.endswith("csv"):
            vals = drycal.sep(fn, ",", encoding, timezone)
        elif filetype.endswith(".txt") or fn.endswith("txt"):
            vals = drycal.sep(fn, "\t", encoding, timezone)

        # check timestamps are increasing:
        warn = True
        ndays = 0
        utslist = vals.uts.values
        for i in range(1, vals.uts.size):
            if utslist[i] < utslist[i - 1]:
                if warn:
                    logger.warning("DryCal log crossing day boundary. Adding offset.")
                    warn = False
                uts = utslist[i] + ndays * 86400
                while uts < utslist[i - 1]:
                    ndays += 1
                    uts = utslist[i] + ndays * 86400
                utslist[i] = uts
        vals["uts"] = xr.DataArray(data=utslist, dims=["uts"])
        vals.attrs["fulldate"] = False
    return vals



"""
**agilentcsv**: Processing Agilent Chemstation Chromtab tabulated data files (csv).
-----------------------------------------------------------------------------------

This file format may include multiple timesteps consisting of several traces each in a
single CSV file. It contains a header section for each timestep, followed by a detector
name, and a sequence of "X, Y" datapoints, which are stored as ``elution_time`` and
``signal``.

.. warning ::

    It is not guaranteed that the X-axis of the chromatogram (i.e. ``elution_time``) is
    consistent between the timesteps of the same trace. The traces are expanded to the
    length of the longest trace, and the shorter traces are padded with ``NaNs``.

.. warning ::

    Unfortunately, the chromatographic ``method`` is not exposed in this file format.

.. codeauthor:: Peter Kraus
"""

import numpy as np
from zoneinfo import ZoneInfo
from uncertainties.core import str_to_number_with_uncert as tuple_fromstr
from ...dgutils.dateutils import str_to_uts
import xarray as xr
from datatree import DataTree


def _process_headers(headers: list, columns: list, timezone: ZoneInfo) -> dict:
    res = {}
    assert len(headers) == len(
        columns
    ), "chromtab: The number of headers and columns do not match."
    assert "Date Acquired" in headers, "chromtab: Cannot infer date."
    res["uts"] = str_to_uts(
        timestamp=columns[headers.index("Date Acquired")].strip(),
        format="%d %b %Y %H:%M",
        timezone=timezone,
    )
    fn = ""
    if "Path" in headers:
        fn += columns[headers.index("Path")]
    if "File" in headers:
        fn += columns[headers.index("File")]
    res["datafile"] = fn
    if "Sample" in headers:
        res["sampleid"] = columns[headers.index("Sample")]
    return res


def _to_trace(tx, ty):
    tvals, tders = [x for x in zip(*tx)]
    yvals, yders = [x for x in zip(*ty)]
    trace = {
        "tvals": np.array(tvals) * 60,
        "tdevs": np.array(tders) * 60,
        "yvals": list(yvals),
        "ydevs": list(yders),
    }
    return trace


def process(*, fn: str, encoding: str, timezone: str, **kwargs: dict) -> DataTree:
    """
    Agilent Chemstation CSV (Chromtab) file parser

    Each file may contain multiple chromatograms per file with multiple traces. Each
    chromatogram starts with a header section, and is followed by each trace, which
    includes a header line and x,y-data.

    Parameters
    ----------
    fn
        Filename to process.

    encoding
        Encoding used to open the file.

    timezone
        Timezone information. This should be ``"localtime"``.

    Returns
    -------
    class:`datatree.DataTree`
        A :class:`datatree.DataTree` containing one :class:`xarray.Dataset` per detector. As
        When multiple timesteps are present in the file, the traces of each detector are
        expanded to match the longest trace, and collated along the ``uts``-dimension.
    """

    with open(fn, "r", encoding=encoding, errors="ignore") as infile:
        lines = infile.readlines()
    metadata = {}
    uts = []
    tx = []
    ty = []
    detname = None
    tstep = dict()
    data = []
    traces = set()
    maxlen = dict()
    for line in lines:
        parts = line.strip().split(",")
        if len(parts) > 2:
            if '"Date Acquired"' in parts:
                if tx != [] and ty != [] and detname is not None:
                    trace = _to_trace(tx, ty)
                    tstep[detname] = trace
                    maxlen[detname] = max(maxlen.get(detname, 0), len(trace["tvals"]))
                    tx = []
                    ty = []
                if len(tstep) > 0:
                    data.append(tstep)
                    tstep = dict()
                headers = [p.replace('"', "") for p in parts]
            else:
                columns = [p.replace('"', "") for p in parts]
                ret = _process_headers(headers, columns, timezone)
                uts.append(ret.pop("uts"))
                metadata.update(ret)
        elif len(parts) == 1:
            if tx != [] and ty != [] and detname is not None:
                trace = _to_trace(tx, ty)
                tstep[detname] = trace
                maxlen[detname] = max(maxlen.get(detname, 0), len(trace["tvals"]))
                tx = []
                ty = []
            detname = parts[0].replace('"', "").split("\\")[-1]
            traces.add(detname)
        elif len(parts) == 2:
            x, y = [tuple_fromstr(i) for i in parts]
            tx.append(x)
            ty.append(y)
    trace = _to_trace(tx, ty)
    tstep[detname] = trace
    maxlen[detname] = max(maxlen.get(detname, 0), len(trace["tvals"]))
    data.append(tstep)

    traces = sorted(traces)
    vals = {}
    for tr in traces:
        dsets = []
        for ti, ts in enumerate(data):
            thislen = len(ts[tr]["tvals"])
            fvals = {}
            for k in {"yvals", "ydevs", "tvals", "tdevs"}:
                fvals[k] = np.ones(maxlen[tr]) * np.nan
                fvals[k][:thislen] = ts[tr][k]
            ds = xr.Dataset(
                data_vars={
                    "signal": (
                        ["elution_time"],
                        fvals["yvals"],
                        {"ancillary_variables": "signal_std_err"},
                    ),
                    "signal_std_err": (
                        ["elution_time"],
                        fvals["ydevs"],
                        {"standard_name": "signal standard_error"},
                    ),
                    "elution_time": (
                        ["_"],
                        fvals["tvals"],
                        {"units": "s", "ancillary_variables": "elution_time_std_err"},
                    ),
                    "elution_time_std_err": (
                        ["elution_time"],
                        fvals["tdevs"],
                        {"units": "s", "standard_name": "elution_time standard_error"},
                    ),
                },
                coords={},
                attrs={},
            )
            ds["uts"] = [uts[ti]]
            dsets.append(ds)
        vals[tr] = xr.concat(dsets, dim="uts")
    dt = DataTree.from_dict(vals)
    dt.attrs = metadata
    return dt



import pytest
from tests.utils import (
    datagram_from_input,
    compare_datatrees,
)
from datatree import open_datatree


@pytest.mark.parametrize(
    "input",
    [
        {  # ts0
            "case": "sheet.UK.tsv",
            "locale": "en_GB.UTF-8",
            "filetype": "None",
            "parameters": {
                "sep": "\t",
                "timestamp": {
                    "date": {"index": 0, "format": "%d/%m/%Y"},
                    "time": {"index": 1, "format": "%H:%M:%S"},
                },
            },
        },
        {  # ts1
            "case": "sheet.DE.tsv",
            "locale": "de_DE.UTF-8",
            "filetype": "None",
            "parameters": {
                "sep": "\t",
                "timestamp": {
                    "date": {"index": 0, "format": "%d.%m.%Y"},
                    "time": {"index": 1, "format": "%H:%M:%S"},
                },
            },
        },
        {  # ts2
            "case": "sheet.US.tsv",
            "locale": "en_US.UTF-8",
            "filetype": "None",
            "parameters": {
                "sep": "\t",
                "timestamp": {
                    "date": {"index": 0, "format": "%m/%d/%Y"},
                    "time": {"index": 1, "format": "%H:%M:%S %p"},
                },
            },
        },
    ],
)
def test_locale_from_basiccsv(input, datadir):
    ret = datagram_from_input(input, "basiccsv", datadir, version="5.0")
    # ret.to_netcdf("ref.sheet.nc")
    print(f"{ret=}")
    ref = open_datatree("ref.sheet.nc")
    compare_datatrees(ret, ref)



import pytest
import os
from yadg.extractors import extract
import datatree
from .utils import compare_datatrees


@pytest.mark.parametrize(
    "filetype, infile, outfile",
    [
        ("marda:biologic-mpr", "cp.mpr", "ref.cp.mpr.nc"),
        ("marda:biologic-mpt", "cp.mpt", "ref.cp.mpt.nc"),
        ("marda:agilent-ch", "hplc.CH", "ref.hplc.ch.nc"),
        ("marda:agilent-dx", "hplc.dx", "ref.hplc.dx.nc"),
        ("marda:phi-spe", "xps.spe", "ref.xps.spe.nc"),
        ("marda:panalytical-xrdml", "xrd.xrdml", "ref.xrd.xrdml.nc"),
    ],
)
def test_extract_marda(filetype, infile, outfile, datadir):
    os.chdir(datadir)
    ret = extract(filetype=filetype, path=infile)
    # ret.to_netcdf(outfile, engine="h5netcdf")
    ref = datatree.open_datatree(outfile)
    print(f"{ret=}")
    compare_datatrees(ret, ref)



"""
**fusionzip**: Processing Inficon Fusion zipped data format (zip).
------------------------------------------------------------------

This is a wrapper parser which unzips the provided zip file, and then uses
the :mod:`yadg.parsers.chromtrace.fusionjson` parser to parse every data
file present in the archive.

Exposed metadata:
`````````````````

.. code-block:: yaml

    method:   !!str
    sampleid: !!str
    version:  !!str
    datafile: !!str

.. codeauthor:: Peter Kraus
"""

import zipfile
import tempfile
import os
import xarray as xr
from datatree import DataTree

from .fusionjson import process as processjson


def process(*, fn: str, encoding: str, timezone: str, **kwargs: dict) -> DataTree:
    """
    Fusion zip file format.

    The Fusion GC's can export their json formats as a zip archive of a folder
    of jsons. This parser allows for parsing of this zip archive directly,
    without the user having to unzip & move the data.

    Parameters
    ----------
    fn
        Filename to process.

    encoding
        Not used as the file is binary.

    timezone
        Timezone information. This should be ``"localtime"``.

    Returns
    -------
    class:`datatree.DataTree`
        A :class:`datatree.DataTree` containing one :class:`xarray.Dataset` per detector. If
        multiple timesteps are found in the zip archive, the :class:`datatree.DataTrees`
        are collated along the ``uts`` dimension.

    """

    zf = zipfile.ZipFile(fn)
    with tempfile.TemporaryDirectory() as tempdir:
        zf.extractall(tempdir)
        dt = None
        for ffn in sorted(os.listdir(tempdir)):
            path = os.path.join(tempdir, ffn)
            if ffn.endswith("fusion-data"):
                fdt = processjson(fn=path, encoding=encoding, timezone=timezone)
                if dt is None:
                    dt = fdt
                elif isinstance(dt, DataTree):
                    for k, v in fdt.items():
                        if k in dt:  # pylint: disable=E1135
                            newv = xr.concat(
                                [dt[k].ds, v.ds],  # pylint: disable=E1136
                                dim="uts",
                                combine_attrs="identical",
                            )
                        else:
                            newv = v.ds
                        dt[k] = DataTree(newv)  # pylint: disable=E1137
                else:
                    raise RuntimeError("We should not get here.")
    return dt



"""
**agilentch**: Processing Agilent OpenLab data archive files (DX).
------------------------------------------------------------------

This is a wrapper parser which unzips the provided DX file, and then uses the
:mod:`yadg.parsers.chromtrace.agilentch` parser to parse every CH file present in
the archive. The IT files in the archive are currently ignored.

In addition to the metadata exposed by the CH parser, the ``datafile`` entry
is populated with the corresponding name of the CH file. The ``fn`` entry in each
timestep contains the parent DX file.

.. note::

   Currently the timesteps from multiple CH files (if present) are appended in the
   timesteps array without any further sorting.

.. codeauthor:: Peter Kraus
"""

import zipfile
import tempfile
import os
from .agilentch import process as processch
from datatree import DataTree
import xarray as xr


def process(*, fn: str, encoding: str, timezone: str, **kwargs: dict) -> DataTree:
    """
    Agilent OpenLab DX archive parser.

    This is a simple wrapper around the Agilent OpenLab signal trace parser in
    :mod:`yadg.parsers.chromtrace.agilentch`. This wrapper first un-zips the DX
    file into a temporary directory, and then processess all CH files found
    within the archive, concatenating timesteps from multiple files.

    Parameters
    ----------
    fn
        Filename to process.

    encoding
        Not used as the file is binary.

    timezone
        Timezone information. This should be ``"localtime"``.

    Returns
    -------
    class:`datatree.DataTree`
        A :class:`datatree.DataTree` containing one :class:`xarray.Dataset` per detector. If
        multiple timesteps are found in the zip archive, the :class:`datatree.DataTrees`
        are collated along the ``uts`` dimension.

    """

    zf = zipfile.ZipFile(fn)
    with tempfile.TemporaryDirectory() as tempdir:
        zf.extractall(tempdir)
        dt = None
        for ffn in os.listdir(tempdir):
            if ffn.endswith("CH"):
                path = os.path.join(tempdir, ffn)
                fdt = processch(fn=path, encoding=encoding, timezone=timezone)
                if dt is None:
                    dt = fdt
                elif isinstance(dt, DataTree):
                    for k, v in fdt.items():
                        if k in dt:  # pylint: disable=E1135
                            try:
                                newv = xr.concat(
                                    [dt[k].ds, v.ds],  # pylint: disable=E1136
                                    dim="uts",
                                    combine_attrs="identical",
                                )
                            except xr.MergeError:
                                raise RuntimeError(
                                    "Merging metadata from the unzipped agilent-ch files has failed. "
                                    "This is a bug. Please open an issue on GitHub."
                                )
                        else:
                            newv = v.ds
                        dt[k] = DataTree(newv)  # pylint: disable=E1137
                else:
                    raise RuntimeError("We should not get here.")
    return dt



import importlib
import logging
from dgbowl_schemas.yadg.dataschema import ExtractorFactory
from pydantic import ValidationError
from pydantic.v1 import ValidationError as ValidationError_v1
from pathlib import Path
from zoneinfo import ZoneInfo
import xarray as xr
import datatree
from typing import Union
from yadg import dgutils, core

logger = logging.getLogger(__name__)

extractors = {}

for modname in {
    "eclabmpr",
    "eclabmpt",
    "agilentch",
    "agilentdx",
    "phispe",
    "panalyticalxrdml",
}:
    try:
        m = importlib.import_module(f"yadg.extractors.{modname}")
        supp = getattr(m, "supports")
        func = getattr(m, "extract")
        for k in supp:
            extractors[k] = func
    except ImportError as e:
        logger.critical(f"could not import module '{modname}'")
        raise e


def extract(filetype: str, path: Path) -> Union[xr.Dataset, datatree.DataTree]:
    """
    Worker function of the ``extract`` subcommand.

    Extracts data from provided ``path``, assuming it is the specified ``filetype``. The
    data is either returned as a :class:`datatree.DataTree` or a :class:`xarray.Dataset`,
    however in either case the returned objects have a :func:`ret.to_netcdf()` as well
    as a :func:`ret.to_dict()` method, which can be used to write the file.

    .. warning::

        This feature has been introduced in ``yadg-5.0a2`` and is unstable.

    Parameters
    ----------

    filetype:
        Specifies the filetype. Can be both ``yadg``-native as well as a MaRDA extractors
        :class:`FileType`. Automatically checks for matches using the ``marda:`` prefix.

    path:
        A :class:`pathlib.Path` object pointing to the file to be extracted.

    Returns
    -------
    Union[xr.Dataset, datatree.DataTree]
        The extracted data and metadata, which can be written to ``NetCDF`` format using
        :func:`ret.to_netcdf()`.

    """
    for k in {filetype, f"marda:{filetype}"}:
        try:
            ftype = ExtractorFactory(extractor={"filetype": k}).extractor
            break
        except (ValidationError, ValidationError_v1) as e:
            logging.debug(e)
            pass
    else:
        raise RuntimeError(f"Filetype '{filetype}' could not be understood.")

    if ftype.filetype in extractors:
        ret = extractors[ftype.filetype](
            fn=str(path),
            encoding=ftype.encoding,
            timezone=ZoneInfo(ftype.timezone),
            locale=ftype.locale,
        )
        ret.attrs = {
            "provenance": "yadg extract",
            "date": dgutils.now(asstr=True),
            "datagram_version": core.datagram_version,
            "yadg_extract_filename": str(path),
            "yadg_extract_filetype": str(ftype),
        }

        return ret


__all__ = ["extract"]



import os
import argparse
import logging
import json
import yaml
import shutil
import hashlib
from pathlib import Path

# from dgbowl_schemas import to_dataschema
from dgbowl_schemas.yadg import to_dataschema
from . import core, dgutils, extractors


logger = logging.getLogger(__name__)


def _load_file(infile: str) -> dict:
    with open(infile, "r") as inf:
        if infile.endswith("json"):
            obj = json.load(inf)
        elif infile.endswith("yml") or infile.endswith("yaml"):
            obj = yaml.safe_load(inf)
        else:
            logging.critical("Filename type not recognised: '%s'", infile)
            raise RuntimeError(f"Filename type not recognised: '{infile}'")
    return obj


def _zip_file(folder: str, outpath: str, method: str = "zip") -> str:
    if method in {"zip", "tar"}:
        fn = f"{outpath}.{method}"
    elif method in {"bztar", "gztar", "xztar"}:
        fn = f"{outpath}.tar.{method[:2]}"
        if method == "bztar":
            fn = f"{fn}2"
    logger.debug("Archiving.")
    shutil.make_archive(outpath, method, root_dir=folder)
    logger.debug("Hashing.")
    m = hashlib.sha1()
    with open(fn, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            m.update(chunk)
    return fn, m.hexdigest()


def process(args: argparse.Namespace) -> None:
    """
    The ``process`` subcommand of **yadg**.

    This function first checks that the supplied ``args.infile`` exists, is a valid
    `schema`, and if yes, proceeds to process the `schema` into a `datagram`. If
    this is successful, the `datagram` is checked for validity and written out
    into ``args.outfile`` (which is `"datagram.json"` by default).
    """
    assert os.path.exists(args.infile) and os.path.isfile(args.infile), (
        f"Supplied schema filename '{args.infile}' does not exist "
        "or is not a valid file."
    )

    logger.info("Reading input file from '%s'.", args.infile)
    schema = _load_file(args.infile)

    logger.info("Loading dataschema.")
    ds = to_dataschema(**schema)
    logger.info("Loaded dataschema version '%s'", ds.metadata.version)

    while hasattr(ds, "update"):
        ds = ds.update()
        if hasattr(ds, "metadata"):
            if hasattr(ds.metadata, "version"):
                if ds.metadata.version == "5.0":
                    break

    logger.debug("Processing schema")
    datagram = core.process_schema(ds, strict_merge=not args.ignore_merge_errors)

    logger.info("Saving datagram to '%s'.", args.outfile)
    datagram.to_netcdf(args.outfile, engine="h5netcdf")
    # with open(args.outfile, "w") as ofile:
    #    json.dump(datagram, ofile, indent=1)


def update(args: argparse.Namespace) -> None:
    """
    The ``update`` subcommand of **yadg**.

    This function updates the `DataSchema` present in the ``args.infile`` argument to
    comply with the newest version of `DataSchema`, and saves the resulting object into
    ``args.outfile`` (which is the ``args.infile`` with  a `".new.json"` suffix
    by default).
    """

    assert os.path.exists(args.infile) and os.path.isfile(args.infile), (
        f"Supplied object filename '{args.infile}' does not exist "
        "or is not a valid file."
    )

    logger.info("Reading input file from '%s'.", args.infile)
    inobj = _load_file(args.infile)

    if args.outfile is None:
        name, ext = os.path.splitext(args.infile)
        args.outfile = f"{name}.new.json"

    outobj = dgutils.update_schema(inobj)

    logger.info("Writing new object into '%s'.", args.outfile)
    with open(args.outfile, "w") as outfile:
        json.dump(outobj.dict(), outfile, indent=1)


def preset(args: argparse.Namespace) -> None:
    """
    The ``preset`` subcommand of **yadg**.

    This function requires the ``args.preset`` and ``args.folder`` arguments.
    If ``args.folder`` is an existing folder and ``args.preset`` a valid `schema`
    file (with appropriately formated ``"import"`` and ``"calfile"`` entries),
    this function will prepend the specified ``args.folder`` to all paths in the
    ``args.preset``, convert them to absolute paths, and save the resulting `schema`
    in the supplied ``args.outfile``.

    Alternatively, if ``args.process`` is specified, the created `schema` will be
    directly processed into a `datagram`, which is then saved in ``args.outfile``.
    """
    assert os.path.exists(args.folder) and os.path.isdir(args.folder), (
        f"Supplied folder path '{args.folder}' does not exist "
        "or is not a valid folder."
    )

    if not os.path.isabs(args.folder) and not args.process:
        logger.warning(
            "The provided path '%s' is a relative path. The generated schema "
            "likely will not work outside current working directory.",
            args.folder,
        )

    assert os.path.exists(args.preset) and os.path.isfile(args.preset), (
        f"Supplied preset path '{args.preset}' does not exist "
        "or is not a valid file."
    )

    logger.info("Reading input file from '%s'.", args.preset)
    preset = _load_file(args.preset)

    logger.info("Processing preset as schema.")
    schema = to_dataschema(**preset)

    logger.info("Creating a schema from preset for '%s'.", args.folder)
    ds = dgutils.schema_from_preset(schema, args.folder)

    logger.info("Loaded dataschema version '%s'", ds.metadata.version)
    if args.process:
        logger.info("Processing created schema.")
        datagram = core.process_schema(ds, strict_merge=not args.ignore_merge_errors)
        args.outfile = "datagram.nc" if args.outfile is None else args.outfile
        if args.archive:
            zipfile = args.outfile.replace(".nc", "")
            logger.info("Zipping input folder into '%s'", zipfile)
            fn, hash = _zip_file(args.folder, zipfile, method=args.packwith)
            datagram.attrs["data_archive_sha-1"] = hash
            datagram.attrs["data_archive_path"] = fn
        logger.info("Saving datagram to '%s'.", args.outfile)
        datagram.to_netcdf(args.outfile, engine="h5netcdf")
    else:
        if args.archive:
            logger.warning(
                "The --archive option must be supplied along with --process. "
                "Continuing without archiving."
            )
        args.outfile = "schema.json" if args.outfile is None else args.outfile
        logger.info("Saving schema to '%s'.", args.outfile)
        with open(args.outfile, "w") as ofile:
            json.dump(ds.dict(), ofile, indent=1)


def extract(args: argparse.Namespace) -> None:
    """
    The ``extract`` subcommand of **yadg**.

    This function requires the ``args.filetype`` and ``args.infile`` positional arguments.
    If ``args.filetype`` (or it's namespaced version, such as ``marda:{args.filetype}``)
    is a known :class:`FileType`, ``yadg`` will attempt to extract metadata and data from
    the provided ``args.infile``.

    The data is returned as a NetCDF file. The location can be configured using the
    ``args.outfile`` parameter, by default this is set to the stem of ``args.infile`` with
    a ``.nc`` suffix.
    """

    path = Path(args.infile)

    assert path.is_file(), (
        f"Supplied object filename '{args.infile}' does not exist "
        "or is not a valid file."
    )

    if args.outfile is None:
        outpath = path.with_suffix(".json" if args.meta_only else ".nc")
    else:
        outpath = Path(args.outfile)

    ret = extractors.extract(args.filetype, path)
    if args.meta_only:
        with outpath.open("w", encoding="UTF-8") as target:
            json.dump(ret.to_dict(data=False), target)
    else:
        ret.to_netcdf(outpath, engine="h5netcdf")



import json
import pytest

from .utils import (
    datagram_from_input,
    standard_datagram_test,
    compare_result_dicts,
    dg_get_quantity,
)


@pytest.mark.parametrize(
    "input, ts",
    [
        (  # ts0 - xy data
            {
                "case": "210520step1_30min.xy",
                "parameters": {"filetype": "panalytical.xy"},
            },
            {"nsteps": 1, "step": 0, "nrows": 1, "point": 0},
        ),
        (  # ts1 - csv data
            {
                "case": "210520step1_30min.csv",
                "parameters": {},
            },
            {"nsteps": 1, "step": 0, "nrows": 1, "point": 0},
        ),
        (  # ts2 - xml data
            {
                "case": "210520step1_30min.xrdml",
                "parameters": {"filetype": "panalytical.xrdml"},
            },
            {"nsteps": 1, "step": 0, "nrows": 1, "point": 0},
        ),
    ],
)
def test_datagram_from_xrdtrace(input, ts, datadir):
    dg = datagram_from_input(input, "xrdtrace", datadir, version="4.1")
    standard_datagram_test(dg, ts)
    with open("xrd_data.json", "r") as inf:
        ref = json.load(inf)
    for k in {"intensity", "angle"}:
        ret = dg_get_quantity(dg, ts["step"], col=k, utsrow=ts["point"])
        compare_result_dicts(ret, ref[k])



import pytest
import os
import json
from tests.utils import standard_datagram_test
from dgbowl_schemas.yadg import to_dataschema

import yadg.dgutils
import yadg.core


@pytest.mark.parametrize(
    "input, ts",
    [
        ("data_1", {"nsteps": 2, "step": 0, "nrows": 1662}),
        ("data_2", {"nsteps": 2, "step": 0, "nrows": 2}),
    ],
)
def test_preset(input, ts, datadir):
    os.chdir(datadir)

    with open(f"{input}.preset.json") as infile:
        preset = json.load(infile)

    schema = to_dataschema(**preset)
    while hasattr(schema, "update"):
        schema = schema.update()
    ds = yadg.dgutils.schema_from_preset(schema, input)

    ret = yadg.core.process_schema(ds)
    standard_datagram_test(ret, ts)



"""
The module handles the reading and processing of the network analyzer
traces, containing the reflection coefficient as a function of the sweeped frequency,
:math:`\\Gamma(f)`.

:mod:`~yadg.parsers.qftrace` loads the reflection trace data, determines the
uncertainties of the signal (y-axis), and explicitly populates the points in
the time axis (x-axis).

Usage
`````
Available since ``yadg-3.0``. The parser supports the following parameters:

.. _yadg.parsers.qftrace.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_0.step.QFTrace

.. _yadg.parsers.qftrace.formats:

 - LabView output in a tab-separated format (``csv``):
   :mod:`~yadg.parsers.qftrace.labviewcsv`

.. _yadg.parsers.qftrace.provides:

Schema
``````
For filetypes containing the reflection trace data, the schema is as follows:

.. code-block:: yaml

  datatree.DataTree:
    S11:
      coords:
        uts:        !!float
        freq:       !!float       # Field frequency (Hz)
      data_vars:
        Re(G):      (uts, freq)   # Imaginary part of the reflection coefficient
        Im(G)       (uts, freq)   # Real part of the reflection coefficient
        average:    (uts)         # Number of scans averaged to form a single trace
        bandwidth:  (uts)         # Filter bandwidth (Hz)

Module Functions
````````````````

"""

from . import labviewcsv
import datatree


def process(
    *,
    filetype: str,
    **kwargs: dict,
) -> datatree.DataTree:
    """
    VNA reflection trace parser. Forwards ``kwargs`` to the worker functions
    based on the supplied ``filetype``.

    Parameters
    ----------
    filetype
        Discriminator used to select the appropriate worker function.

    Returns
    -------
    :class:`datatree.DataTree`

    """
    if filetype == "labview.csv":
        return labviewcsv.process(**kwargs)



"""
This parser handles the reading and processing of the legacy log files created by
the LabView interface for the MCPT instrument. These files contain information about
the timestamp, temperatures, and inlet / process flows.

.. admonition:: DEPRECATED in ``yadg-4.0``

    As of ``yadg-4.0``, this parser is deprecated and should not be used for new data.
    Please consider switching to the :mod:`~yadg.parsers.basiccsv` parser.

Usage
`````
Available since ``yadg-3.0``. Deprecated since ``yadg-4.0``. The parser supports the
following parameters:

.. _yadg.parsers.meascsv.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_0.step.MeasCSV

.. _parsers_meascsv_provides:

Schema
``````
The parser is used to extract all of the tabular data in the input file, storing
them in the same format as :mod:`~yadg.parsers.basiccsv`, using the column headers
as keys.

"""

import logging
from pydantic import BaseModel
from zoneinfo import ZoneInfo
from ..basiccsv.main import process_row, append_dicts, dicts_to_dataset
from ... import dgutils
import xarray as xr

logger = logging.getLogger(__name__)


def process(
    *,
    fn: str,
    encoding: str,
    timezone: ZoneInfo,
    parameters: BaseModel,
    **kwargs: dict,
) -> xr.Dataset:
    """
    Legacy MCPT measurement log parser.

    This parser is included to maintain parity with older schemas and datagrams.
    It is essentially a wrapper around :func:`yadg.parsers.basiccsv.main.process_row`.

    .. admonition:: DEPRECATED in ``yadg-4.0``

        For new applications, please use the :mod:`~yadg.parsers.basiccsv` parser.

    Parameters
    ----------
    fn
        File to process

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "localtime".

    parameters
        Parameters for :class:`~dgbowl_schemas.yadg.dataschema_5_0.step.MeasCSV`.

    Returns
    -------
    :class:`xarray.Dataset`
        A :class:`xarray.Dataset` containing the timesteps, metadata, and full date tag. No
        metadata is returned. The full date is always provided in :mod:`~yadg.parsers.meascsv`
        compatible files.

    """
    logger.warning("This parser is deprecated. Please switch to 'basiccsv'.")

    with open(fn, "r", encoding=encoding) as infile:
        lines = [i.strip() for i in infile.readlines()]

    headers = [i.strip() for i in lines.pop(0).split(";")]

    for hi, header in enumerate(headers):
        if "/" in header:
            logger.warning("Replacing '/' for '_' in header '%s'.", header)
            headers[hi] = header.replace("/", "_")

    _units = [i.strip() for i in lines.pop(0).split(";")]
    units = {}
    for h in headers:
        units[h] = _units.pop(0)

    units = dgutils.sanitize_units(units)

    datecolumns, datefunc, fulldate = dgutils.infer_timestamp_from(
        spec=parameters.timestamp,
        timezone=timezone,
    )

    # Process rows
    data_vals = {}
    meta_vals = {"_fn": []}
    for li, line in enumerate(lines):
        vals, devs = process_row(
            headers,
            line.split(";"),
            datefunc,
            datecolumns,
        )
        append_dicts(vals, devs, data_vals, meta_vals, fn, li)

    return dicts_to_dataset(data_vals, meta_vals, units, fulldate)



import pytest
import os
import json
from tests.utils import (
    datagram_from_input,
    standard_datagram_test,
    compare_result_dicts,
    dg_get_quantity,
)


@pytest.mark.parametrize(
    "input, ts",
    [
        (
            {  # ts0 - datasc parse & integration from file
                "folders": ["."],
                "prefix": "2019-12-03",
                "suffix": "dat.asc",
                "parameters": {"tracetype": "ezchrom.asc"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 5,
                "method": "polyarc+TCD_PK_09b.met",
                "point": 1,
                "test": {
                    "0": {
                        "i": 1000,
                        "elution_time": {"n": 100.002, "s": 0.100002, "u": "s"},
                        "signal": {"n": 2.085281, "s": 0.000130, "u": "pA"},
                    },
                    "1": {
                        "i": 1000,
                        "elution_time": {"n": 100.002, "s": 0.100002, "u": "s"},
                        "signal": {"n": 55.892435, "s": 0.000130, "u": "V"},
                    },
                },
            },
        ),
        (
            {  # ts1 - chromtab parse
                "folders": ["."],
                "prefix": "CHROMTAB",
                "suffix": "CSV",
                "parameters": {"tracetype": "agilent.csv"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 3,
                "method": None,
                "point": 0,
                "test": {
                    "TCD1A.ch": {
                        "i": 10,
                        "elution_time": {"n": 2.15999999, "s": 0.06, "u": "s"},
                        "signal": {"n": 488830.0, "s": 0.001, "u": None},
                    },
                    "FID2B.ch": {
                        "i": 10,
                        "elution_time": {"n": 2.15999999, "s": 0.06, "u": "s"},
                        "signal": {"n": 42430.0, "s": 0.001},
                    },
                },
            },
        ),
        (
            {  # ts2 - fusion json parse
                "folders": ["."],
                "prefix": "",
                "suffix": "fusion-data",
                "parameters": {"tracetype": "fusion.json"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 5,
                "method": "AS_Cal_20210702",
                "point": 4,
                "test": {
                    "moduleA:tcd": {
                        "i": 100,
                        "elution_time": {"n": 2.0, "s": 0.02, "u": "s"},
                        "signal": {"n": 1566.0, "s": 1.0, "u": None},
                    },
                    "moduleB:tcd": {
                        "i": 100,
                        "elution_time": {"n": 2.0, "s": 0.02, "u": "s"},
                        "signal": {"n": -1412.0, "s": 1.0},
                    },
                },
            },
        ),
        (
            {  # ts3 - fusion zip parse
                "folders": ["."],
                "prefix": "",
                "suffix": "zip",
                "parameters": {"tracetype": "fusion.zip"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 12,
                "method": "AS_Cal_20220204",
                "point": 4,
                "test": {
                    "moduleA:tcd": {
                        "i": 0,
                        "elution_time": {"n": 0.0, "s": 0.02, "u": "s"},
                        "signal": {"n": 0.0, "s": 1.0, "u": None},
                    }
                },
            },
        ),
        (
            {  # ts4 - agilent ch file parse
                "folders": ["."],
                "suffix": "CH",
                "parameters": {"tracetype": "agilent.ch"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 1,
                "method": "CO2RR_FTI_0.7mL_40uL_45dgr_36min.amx",
                "point": 0,
                "test": {
                    "RID1A": {
                        "i": 0,
                        "elution_time": {"n": 0.0675, "s": 0.0675, "u": "s"},
                        "signal": {"n": 0.39, "s": 0.01, "u": "nRIU"},
                    }
                },
            },
        ),
        (
            {  # ts5 - agilent dx file unzip & parse
                "folders": ["."],
                "suffix": "dx",
                "parameters": {"tracetype": "agilent.dx"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 1,
                "method": "CO2RR_FTI_0.7mL_40uL_45dgr_36min.amx",
                "point": 0,
                "test": {
                    "RID1A": {
                        "i": 10,
                        "elution_time": {"n": 2.2276485, "s": 0.0675, "u": "s"},
                        "signal": {"n": -1.6, "s": 0.01, "u": "nRIU"},
                    }
                },
            },
        ),
    ],
)
def test_datagram_from_chromtrace(input, ts, datadir):
    os.chdir(datadir)
    dg = datagram_from_input(input, "chromtrace", datadir)
    standard_datagram_test(dg, ts)
    if isinstance(ts["step"], str):
        step = dg[ts["step"]]
    else:
        step = dg[list(dg.children.keys())[ts["step"]]]
    for k, v in ts["test"].items():
        for kk in {"signal", "elution_time"}:
            ret = dg_get_quantity(step, k, col=kk, utsrow=ts["point"])
            print(f"{v[kk]=}")
            print(f"{ret=}")
            compare_result_dicts(
                {"n": ret["n"][v["i"]], "s": ret["s"][v["i"]], "u": ret["u"]},
                v[kk],
            )


@pytest.mark.parametrize(
    "input",
    [
        (
            {  # ts0 - ch file parse, method, and integration
                "folders": ["."],
                "suffix": "CH",
                "parameters": {"tracetype": "agilent.ch"},
            }
        ),
        (
            {  # ts1 - dx file unzip, parse, method, integration from file
                "folders": ["."],
                "suffix": "dx",
                "parameters": {"tracetype": "agilent.dx"},
            }
        ),
    ],
)
def test_chromtrace_compare_raw_values(input, datadir):
    os.chdir(datadir)
    dg = datagram_from_input(input, "chromtrace", datadir)
    with open("yvals.json", "r") as infile:
        ref = json.load(infile)["traces"]
    for trace, v in ref.items():
        for k in {"signal", "elution_time"}:
            ret = dg_get_quantity(dg["0"], trace, col=k, utsrow=0)
            compare_result_dicts(ret, v[k])



"""
panalyticalxrdml: Processing of PANalytical XRD ``xml`` files
-------------------------------------------------------------

File Structure
``````````````

These are xml-formatted files, which we here parse using the :mod:`xml.etree`
library into a Python :class:`dict`.

.. note::

    The ``angle`` returned from this parser is based on a linear interpolation of
    the start and end point of the scan, and is the :math:`2\\theta`. The values
    of :math:`\\omega` are discarded.

.. warning::

    This parser is fairly new and untested. As a result, the returned metadata
    contain only a subset of the available metadata in the XML file. If something
    important is missing, please contact us!

Uncertainties
`````````````
The uncertainties of in ``"angle"`` are taken as the step-width of
the linearly spaced :math:`2\\theta` values.

The uncertainties of of ``"intensity"`` are currently set to a constant
value of 1.0 count as all the supported files seem to produce integer values.

.. codeauthor::
    Nicolas Vetsch,
    Peter Kraus
"""

from collections import defaultdict
from typing import Union
from xml.etree import ElementTree
import numpy as np
import xarray as xr
from zoneinfo import ZoneInfo

from uncertainties.core import str_to_number_with_uncert as tuple_fromstr
from .common import panalytical_comment
from ...dgutils import dateutils


def etree_to_dict(e: ElementTree.Element) -> dict:
    """Recursively converts an ElementTree.Element into a dictionary.

    Element attributes are stored into `"@"`-prefixed attribute keys.
    Element text is stored into `"#text"` for all nodes.

    From https://stackoverflow.com/a/10076823.

    Parameters
    ----------
    e
        The ElementTree root Element.

    Returns
    -------
    dict
        ElementTree parsed into a dictionary.

    """
    d = {e.tag: {} if e.attrib else None}
    children = list(e)
    if children:
        dd = defaultdict(list)
        for dc in map(etree_to_dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        d = {e.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if e.attrib:
        d[e.tag].update(("@" + k, v) for k, v in e.attrib.items())
    if e.text:
        text = e.text.strip()
        if children or e.attrib:
            if text:
                d[e.tag]["#text"] = text
        else:
            d[e.tag] = text
    return d


def _process_values(d: Union[dict, str]) -> Union[dict, str]:
    """
    Recursively parses dicts in the following format:

    .. code::

        {"key": {"#text": ..., "@unit": ...}, ...}

    into a single string:

    .. code::

        {"key": f"{#text} {@unit}", ...}

    """
    # TODO
    # If not "#text" or @tribute just snake_case and recurse.
    if isinstance(d, dict):
        if "@unit" in d and "#text" in d:
            return f"{d['#text']} {d['@unit']}"
        elif "@version" in d and "#text" in d:
            return f"{d['#text']} {d['@version']}"
        else:
            for k, v in d.items():
                d[k] = _process_values(v)
    return d


def _process_scan(scan: dict) -> dict:
    """
    Parses the scan section of the file. Creates the explicit positions based
    on the number of measured intensities and the start & end position.

    """
    header = scan.pop("header")
    dpts = scan.pop("dataPoints")
    counting_time = _process_values(dpts.pop("commonCountingTime"))
    ivals, idevs = list(
        zip(*[tuple_fromstr(c) for c in dpts["intensities"].pop("#text").split()])
    )
    iunit = dpts["intensities"].pop("@unit")
    timestamp = header.pop("startTimeStamp")

    dp = {
        "intensity": {"vals": ivals, "devs": idevs, "unit": iunit},
        "timestamp": timestamp,
        "counting_time": counting_time,
    }

    positions = _process_values(dpts.pop("positions"))
    for v in positions:
        pos = np.linspace(
            float(v["startPosition"]), float(v["endPosition"]), num=len(ivals)
        )
        adiff = np.abs(np.diff(pos)) * 0.5
        adiff = np.append(adiff, adiff[-1])
        dp[v["@axis"]] = {
            "vals": pos,
            "devs": adiff,
            "unit": v["@unit"],
        }
    return dp


def _process_comment(comment: dict) -> dict:
    """ """
    entry = comment.pop("entry")
    ret = {}
    for line in entry:
        ret.update(panalytical_comment(line))
    return ret


def _process_measurement(measurement: dict, timezone: str):
    """
    A function that processes each section of the XRD XML file.
    """
    # Comment.
    comment = measurement["comment"].pop("entry")
    for line in comment:
        if "PHD Lower Level" in line and "PHD Upper Level" in line:
            __, values = list(zip(*[s.split(" = ") for s in line.split(", ")]))
    keys = ["phd_lower_level", "phd_upper_level"]
    measurement["comment"] = dict(zip(keys, values))
    # Wavelength.
    wavelength = _process_values(measurement.pop("usedWavelength"))
    measurement["wavelength"] = wavelength
    # Incident beam path.
    incident_beam_path = _process_values(measurement.pop("incidentBeamPath"))
    measurement["incident_beam_path"] = incident_beam_path
    # Diffracted beam path.
    diffracted_beam_path = _process_values(measurement.pop("diffractedBeamPath"))
    measurement["diffracted_beam_path"] = diffracted_beam_path
    scan = _process_scan(measurement.pop("scan"))
    trace = {"angle": scan.pop("2Theta"), "intensity": scan.pop("intensity")}
    meta = measurement
    meta["counting_time"] = scan.pop("counting_time")
    trace["uts"] = dateutils.str_to_uts(
        timestamp=scan.pop("timestamp"), timezone=timezone
    )
    return trace, meta


def process(
    *,
    fn: str,
    timezone: ZoneInfo,
    **kwargs: dict,
) -> xr.Dataset:
    """Processes a PANalytical xrdml file.

    Parameters
    ----------
    fn
        The file containing the trace(s) to parse.

    timezone
        A string description of the timezone. Default is "UTC".

    Returns
    -------
    :class:`xarray.Dataset`
        Data containing the timesteps, and metadata. This filetype contains the full
        date specification.

    """
    it = ElementTree.iterparse(fn)
    # Removing xmlns prefixes from all tags.
    # From https://stackoverflow.com/a/25920989.
    for __, e in it:
        __, xmlns_present, postfix = e.tag.partition("}")
        if xmlns_present:
            e.tag = postfix  # Strip away all xmlns prefixes.
    root = it.root
    xrd = etree_to_dict(root)
    # Start processing the xml contents.
    measurements = xrd["xrdMeasurements"]
    assert measurements["@status"] == "Completed", "Incomplete measurement."
    comment = _process_comment(measurements["comment"])
    # Renaming some entries because I want to.
    sample = measurements["sample"]
    sample["prepared_by"] = sample.pop("preparedBy")
    sample["type"] = sample.pop("@type")
    # Process measurement data.
    data, meta = _process_measurement(measurements["xrdMeasurement"], timezone)
    data["fn"] = fn
    # Shove unused data into meta
    meta["sample"] = sample
    meta["comment"] = comment
    meta["fulldate"] = True
    # Build Datasets
    vals = xr.Dataset(
        data_vars={
            "intensity": (
                ["uts", "angle"],
                np.reshape(data["intensity"]["vals"], (1, -1)),
                {
                    "units": data["intensity"]["unit"],
                    "ancillary_variables": "intensity_std_err",
                },
            ),
            "intensity_std_err": (
                ["uts", "angle"],
                np.reshape(data["intensity"]["devs"], (1, -1)),
                {
                    "units": data["intensity"]["unit"],
                    "standard_name": "intensity standard_error",
                },
            ),
            "angle_std_err": (
                ["uts", "angle"],
                np.reshape(data["angle"]["devs"], (1, -1)),
                {
                    "units": data["angle"]["unit"],
                    "standard_name": "angle standard_error",
                },
            ),
        },
        coords={
            "uts": (["uts"], [data["uts"]]),
            "angle": (
                ["angle"],
                data["angle"]["vals"],
                {
                    "units": data["angle"]["unit"],
                    "ancillary_variables": "angle_std_err",
                },
            ),
        },
        attrs=meta,
    )
    return vals



"""
Handles the reading and processing of any tabular files, as long as the first line
contains the column headers. By default, the second should contain the units. The
columns of the table must be separated using a separator such as ``,``, ``;``,
or ``\\t``.

.. warning::

  Since ``yadg-5.0``, the parser handles sparse tables (i.e. tables with missing
  data) by back-filling empty cells with ``np.NaNs``.

.. note::

  :mod:`~yadg.parsers.basiccsv` attempts to deduce the timestamp from the column
  headers, using :func:`yadg.dgutils.dateutils.infer_timestamp_from`. Alternatively,
  the column(s) containing the timestamp data and their format can be provided using
  parameters.

Usage
`````
Available since ``yadg-4.0``. The parser supports the following parameters:

.. _yadg.parsers.basiccsv.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_0.step.BasicCSV

Schema
``````
The primary functionality of :mod:`~yadg.parsers.basiccsv` is to load the tabular
data, and determine the Unix timestamp. The headers of the tabular data are taken
`verbatim` from the file, and appear as ``data_vars`` of the :class:`xarray.Dataset`.
The single ``coord`` for the ``data_vars`` is the deduced Unix timestamp, ``uts``.

.. code-block:: yaml

  xr.Dataset:
    coords:
      uts:            !!float               # Unix timestamp
    data_vars:
      {{ headers }}:  (uts)                 # Populated from file headers

Module Functions
````````````````

"""

from .main import process

__all__ = ["process"]



"""
Handles the reading of post-processed chromatography data, i.e. files containing peak
areas, concentrations, or mole fractions.

.. note::

  To parse trace data as present in raw chromatograms, use the
  :mod:`~yadg.parsers.chromtrace` parser.

Usage
`````
Available since ``yadg-4.2``. The parser supports the following parameters:

.. _yadg.parsers.chromdata.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_0.step.ChromData

.. _yadg.parsers.chromdata.formats:

Formats
```````
The ``filetypes`` currently supported by the parser are:

 - Inficon Fusion JSON format (``fusion.json``):
   see :mod:`~yadg.parsers.chromdata.fusionjson`
 - Inficon Fusion zip archive (``fusion.zip``):
   see :mod:`~yadg.parsers.chromdata.fusionzip`
 - Inficon Fusion csv export (``fusion.csv``):
   see :mod:`~yadg.parsers.chromdata.fusioncsv`
 - Empa's Agilent LC csv export (``empalc.csv``):
   see :mod:`~yadg.parsers.chromdata.empalccsv`
 - Empa's Agilent LC excel export (``empalc.xlsx``):
   see :mod:`~yadg.parsers.chromdata.empalcxlsx`

Schema
``````
Each file is processed into a single :class:`xarray.Dataset`, containing the following
``coords`` and ``data_vars`` (if present in the file):

.. code-block:: yaml

  xr.Dataset:
    coords:
      uts:            !!float               # Unix timestamp
      species:        !!str                 # Species names
    data_vars:
      height:         (uts, species)        # Peak height maximum
      area:           (uts, species)        # Integrated peak area
      retention time: (uts, species)        # Peak retention time
      concentration:  (uts, species)        # Species concentration (mol/l)
      xout:           (uts, species)        # Species mole fraction (-)

Module Functions
````````````````

"""

import xarray as xr

from . import (
    fusionjson,
    fusionzip,
    fusioncsv,
    empalccsv,
    empalcxlsx,
)


def process(*, filetype: str, **kwargs: dict) -> xr.Dataset:
    """
    Unified chromatographic data parser. Forwards ``kwargs`` to the worker functions
    based on the supplied ``filetype``.

    Parameters
    ----------
    filetype
        Discriminator used to select the appropriate worker function.

    Returns
    -------
    :class:`xarray.Dataset`

    """
    if filetype == "fusion.json":
        return fusionjson.process(**kwargs)
    elif filetype == "fusion.zip":
        return fusionzip.process(**kwargs)
    elif filetype == "fusion.csv":
        return fusioncsv.process(**kwargs)
    elif filetype == "empalc.csv":
        return empalccsv.process(**kwargs)
    elif filetype == "empalc.xlsx":
        return empalcxlsx.process(**kwargs)



import pytest
import os
import yaml
from tests.utils import (
    datagram_from_input,
    standard_datagram_test,
    pars_datagram_test,
)
from yadg.core import process_schema
from dgbowl_schemas.yadg import to_dataschema


@pytest.mark.parametrize(
    "input, ts",
    [
        (
            {  # ts0 - rtf parser, supplied date
                "case": "Cp_100mA_1mindelay.rtf",
                "parameters": {"filetype": "drycal.rtf"},
                "externaldate": {"from": {"isostring": "2021-09-17"}},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 110,
                "point": 0,
                "pars": {
                    "Temp": {"sigma": 0.1, "value": 27.4, "unit": "degC"},
                    "uts": {"value": 1631880613.0},
                },
            },
        ),
        (
            {  # ts1 - default sep parser, date from fn
                "case": "20211011_DryCal_out.csv",
                "parameters": {"filetype": "drycal.csv"},
                "externaldate": {"from": {"filename": {"format": "%Y%m%d", "len": 8}}},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 29,
                "point": 0,
                "pars": {
                    "Temp": {"sigma": 0.1, "value": 24.3, "unit": "degC"},
                    "uts": {"value": 1633956333.0},
                },
            },
        ),
        (
            {  # ts2 - default sep parser, date from fn, sigma from length
                "case": "2021-10-11_DryCal_out.txt",
                "parameters": {"filetype": "drycal.txt"},
                "externaldate": {
                    "from": {"filename": {"format": "%Y-%m-%d", "len": 10}}
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 29,
                "point": 28,
                "pars": {
                    "Pressure": {"sigma": 0.001, "value": 971.0, "unit": "mbar"},
                    "uts": {"value": 1633958360.0},
                },
            },
        ),
        (
            {  # ts3 - default sep parser, date from fn
                "case": "2021-10-11_DryCal_out.txt",
                "parameters": {"filetype": "drycal.txt"},
                "externaldate": {
                    "from": {"filename": {"format": "%Y-%m-%d", "len": 10}}
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 29,
                "point": 28,
                "pars": {
                    "DryCal": {"sigma": 0.0001, "value": 14.848, "unit": "smL/min"},
                    "Pressure": {"sigma": 0.001, "value": 971.0, "unit": "mbar"},
                },
            },
        ),
        (
            {  # ts4 - overnight run
                "case": "20220721-porosity-study-20p-Cu-200mA-EDLC-01-flow.csv",
                "parameters": {"filetype": "drycal.csv"},
                "externaldate": {"from": {"filename": {"format": "%Y%m%d", "len": 8}}},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 77,
                "point": 76,
                "pars": {
                    "Temp": {"sigma": 0.1, "value": 28.6, "unit": "degC"},
                    "uts": {"value": 1658475079.0},
                },
            },
        ),
    ],
)
def test_datagram_from_drycal(input, ts, datadir):
    ret = datagram_from_input(input, "flowdata", datadir)
    standard_datagram_test(ret, ts)
    pars_datagram_test(ret, ts)


def test_lock_stock_drycal(datadir):
    os.chdir(datadir)
    with open("lock_stock_dataschema.yml", "r") as inf:
        schema = yaml.safe_load(inf)
    ret = process_schema(to_dataschema(**schema))
    print(f"{ret=}")
    assert ret["outlet"]["DryCal"].shape == (187,)



"""
**ezchromasc**: Processing EZ-Chrom ASCII export files (dat.asc).
-----------------------------------------------------------------

This file format includes one timestep with multiple traces in each ASCII file. It
contains a header section, and a sequence of Y datapoints (``signal``) for each detector.
The X-axis (``elution_time``) is assumed to be uniform between traces, and its units have
to be deduced from the header.

.. codeauthor:: Peter Kraus
"""

import numpy as np
import logging
from zoneinfo import ZoneInfo
from uncertainties.core import str_to_number_with_uncert as tuple_fromstr
from ...dgutils.dateutils import str_to_uts
import xarray as xr
from datatree import DataTree

logger = logging.getLogger(__name__)


def process(*, fn: str, encoding: str, timezone: ZoneInfo, **kwargs: dict) -> DataTree:
    """
    EZ-Chrome ASCII export file parser.

    One chromatogram per file with multiple traces. A header section is followed by
    y-values for each trace. x-values have to be deduced using number of points,
    frequency, and x-multiplier. Method name is available, but detector names are not.
    They are assigned their numerical index in the file.

    Parameters
    ----------
    fn
        Filename to process.

    encoding
        Encoding used to open the file.

    timezone
        Timezone information. This should be ``"localtime"``.


    Returns
    -------
    class:`datatree.DataTree`
        A :class:`datatree.DataTree` containing one :class:`xarray.Dataset` per detector.

    """

    with open(fn, "r", encoding=encoding, errors="ignore") as infile:
        lines = infile.readlines()
    metadata = {}
    data = {}

    for line in lines:
        for key in ["Version", "Method", "User Name"]:
            if line.startswith(key):
                k = key.lower().replace(" ", "")
                metadata[k] = line.split(f"{key}:")[1].strip()
        for key in ["Sample ID"]:  # , "Data File"]:
            if line.startswith(key):
                k = key.lower().replace(" ", "")
                metadata[k] = line.split(f"{key}:")[1].strip()
        if line.startswith("Acquisition Date and Time:"):
            uts = str_to_uts(
                timestamp=line.split("Time:")[1].strip(),
                format="%m/%d/%Y %I:%M:%S %p",
                timezone=timezone,
            )
        if line.startswith("Sampling Rate:"):
            assert (
                "Hz" in line
            ), f"datasc: Incorrect units for rate in file {fn}: {line}"
            parts = line.split("\t")
            samplerates = [float(each.strip()) for each in parts[1:-1]]
        if line.startswith("Total Data Points:"):
            assert (
                "Pts." in line
            ), f"datasc: Incorrect units for number of points in file {fn}: {line}"
            parts = line.split("\t")
            npoints = [int(each.strip()) for each in parts[1:-1]]
        if line.startswith("X Axis Title:"):
            parts = line.split("\t")
            xunits = [each.strip() for each in parts[1:]]
        if line.startswith("Y Axis Title:"):
            parts = line.split("\t")
            yunits = [each.strip() for each in parts[1:]]
            if "25 V" in yunits:
                logger.warning("Implicit conversion of y-axis unit from '25 V' to 'V'.")
                yunits = [i.replace("25 V", "V") for i in yunits]
        if line.startswith("X Axis Multiplier:"):
            parts = line.split("\t")
            xmuls = [float(each.strip()) for each in parts[1:]]
        if line.startswith("Y Axis Multiplier:"):
            parts = line.split("\t")
            ymuls = [float(each.strip()) for each in parts[1:]]
        if ":" not in line:
            si = lines.index(line)
            break
    assert (
        len(samplerates)
        == len(npoints)
        == len(xunits)
        == len(yunits)
        == len(xmuls)
        == len(ymuls)
    ), f"datasc: Inconsistent number of traces in {fn}."

    data = {}
    units = {}
    for ti, npts in enumerate(npoints):
        assert (
            xunits[ti] == "Minutes"
        ), f"datasc: X units label of trace {ti} in {fn} was not understood."
        dt = 60
        xmul = xmuls[ti] * dt / samplerates[ti]
        ymul = ymuls[ti]
        xsn = np.arange(npts) * xmul
        xss = np.ones(npts) * xmul
        ysn, yss = zip(*[tuple_fromstr(li) for li in lines[si : si + npts]])
        si += npts
        data[f"{ti}"] = {
            "t": (xsn, xss),
            "y": (np.array(ysn) * ymul, np.array(yss) * ymul),
        }
        units[f"{ti}"] = {"t": "s", "y": yunits[ti]}

    traces = sorted(data.keys())
    vals = {}
    for ti in traces:
        fvals = xr.Dataset(
            data_vars={
                "signal": (
                    ["uts", "elution_time"],
                    [data[ti]["y"][0]],
                    {"units": units[ti]["y"], "ancillary_variables": "signal_std_err"},
                ),
                "signal_std_err": (
                    ["uts", "elution_time"],
                    [data[ti]["y"][1]],
                    {"units": units[ti]["y"], "standard_name": "signal standard_error"},
                ),
                "elution_time_std_err": (
                    ["elution_time"],
                    data[ti]["t"][1],
                    {
                        "units": units[ti]["t"],
                        "standard_name": "elution_time standard_error",
                    },
                ),
            },
            coords={
                "elution_time": (
                    ["elution_time"],
                    data[ti]["t"][0],
                    {
                        "units": units[ti]["t"],
                        "ancillary_variables": "elution_time_std_err",
                    },
                ),
                "uts": (["uts"], [uts]),
            },
            attrs={},
        )
        vals[ti] = fvals
    dt = DataTree.from_dict(vals)
    dt.attrs = metadata
    return dt



import json
import os

import numpy as np
import pytest
import xarray
from pathlib import Path

import yadg.extractors
from tests.utils import (
    datagram_from_file,
    datagram_from_input,
    pars_datagram_test,
    standard_datagram_test,
    dg_get_quantity,
    compare_result_dicts,
)


@pytest.mark.parametrize(
    "input, ts",
    [
        (  # ts0 - ca.mpr
            {
                "case": "ca.mpr",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpr"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 721,
                "point": 0,
                "pars": {
                    "control_V": {
                        "value": 3.0000624e-001,
                        "sigma": 2e-5,
                        "unit": "V",
                    },
                    "uts": {"value": 1556661461.2284908},
                },
            },
        ),
        (  # ts1 - ca.mpt
            {
                "case": "ca.mpt",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpt"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 721,
                "point": 0,
                "pars": {
                    "control_V": {
                        "value": 3.0000624e-001,
                        "sigma": 2e-5,
                        "unit": "V",
                    },
                    "uts": {"value": 1556661461.2284908},
                },
            },
        ),
        (  # ts2 - cp.mpr
            {
                "case": "cp.mpr",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpr"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 121,
                "point": 0,
                "pars": {
                    "control_I": {
                        "value": -1.0000000e002,
                        "sigma": 4e-5,
                        "unit": "mA",
                    }
                },
            },
        ),
        (  # ts3 - cp.mpt
            {
                "case": "cp.mpt",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpt"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 121,
                "point": 0,
                "pars": {
                    "control_I": {
                        "value": -1.0000000e002,
                        "sigma": 4e-5,
                        "unit": "mA",
                    }
                },
            },
        ),
        (  # ts4 - cv.mpr
            {
                "case": "cv.mpr",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpr"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 5103,
                "point": 0,
                "pars": {
                    "control_V": {
                        "value": 2.4814086e000,
                        "sigma": 2e-4,
                        "unit": "V",
                    }
                },
            },
        ),
        (  # ts5 - cv.mpt
            {
                "case": "cv.mpt",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpt"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 5103,
                "point": 0,
                "pars": {
                    "control_V": {
                        "value": 2.4814086e000,
                        "sigma": 2e-4,
                        "unit": "V",
                    }
                },
            },
        ),
        (  # ts6 - gcpl.mpr
            {
                "case": "gcpl.mpr",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpr"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 12057,
                "point": 0,
                "pars": {
                    "control_I": {
                        "value": -1.1000001e-002,
                        "sigma": 760e-12,
                        "unit": "mA",
                    }
                },
            },
        ),
        (  # ts7 - gcpl.mpt
            {
                "case": "gcpl.mpt",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpt"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 12057,
                "point": 0,
                "pars": {
                    "control_I": {
                        "value": -1.1000001e-002,
                        "sigma": 760e-12,
                        "unit": "mA",
                    }
                },
            },
        ),
        (  # ts8 - geis.mpr
            {
                "case": "geis.mpr",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpr"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 2528,
                "point": 0,
                "pars": {"uts": {"value": 1614849363.9204996}},
            },
        ),
        (  # ts9 - geis.mpt
            {
                "case": "geis.mpt",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpt"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 2528,
                "point": 0,
                "pars": {"uts": {"value": 1614849363.9204996}},
            },
        ),
        (  # ts10 - lsv.mpr
            {
                "case": "lsv.mpr",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpr"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 1186,
                "point": 0,
                "pars": {
                    "control_V": {
                        "value": -5.6856550e-002,
                        "sigma": 3.0518e-4,
                        "unit": "V",
                    }
                },
            },
        ),
        (  # ts11 - lsv.mpt
            {
                "case": "lsv.mpt",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpt"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 1186,
                "point": 0,
                "pars": {
                    "control_V": {
                        "value": -5.6856550e-002,
                        "sigma": 3.0518e-4,
                        "unit": "V",
                    }
                },
            },
        ),
        (  # ts12 - mb.mpr
            {
                "case": "mb.mpr",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpr"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 1588,
                "point": 0,
                "pars": {
                    "control_I": {
                        "value": 1.0857184e-004,
                        "sigma": 4e-9,
                        "unit": "mA",
                    }
                },
            },
        ),
        (  # ts13 - mb.mpt
            {
                "case": "mb.mpt",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpt"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 1588,
                "point": 0,
                "pars": {
                    "control_I": {
                        "value": 1.0857184e-004,
                        "sigma": 4e-9,
                        "unit": "mA",
                    }
                },
            },
        ),
        (  # ts14 - ocv.mpr
            {
                "case": "ocv.mpr",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpr"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 602,
                "point": 0,
                "pars": {"uts": {"value": 1579184014.0}},
            },
        ),
        (  # ts15 - ocv.mpt
            {
                "case": "ocv.mpt",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpt"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 602,
                "point": 0,
                "pars": {"uts": {"value": 1579184014.0}},
            },
        ),
        (  # ts16 - peis.mpr
            {
                "case": "peis.mpr",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpr"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 32,
                "point": 0,
                "pars": {"uts": {"value": 1614702327.1567264}},
            },
        ),
        (  # ts17 - peis.mpt
            {
                "case": "peis.mpt",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpt"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 32,
                "point": 0,
                "pars": {"uts": {"value": 1614702327.1567264}},
            },
        ),
        (  # ts18 - zir.mpr
            {
                "case": "zir.mpr",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpr"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 1,
                "point": 0,
                "pars": {"uts": {"value": 1522109455.9886506}},
            },
        ),
        (  # ts19 - zir.mpt
            {
                "case": "zir.mpt",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpt"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 1,
                "point": 0,
                "pars": {"uts": {"value": 1522109455.9886506}},
            },
        ),
        (  # ts20 - mb_67.mpr
            {
                "case": "mb_67.mpr",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpr"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 33,
                "point": 0,
                "pars": {
                    "uts": {"value": 1670510213.355},
                    "Ewe": {"value": 2.3278546, "unit": "V", "sigma": 7.5e-05},
                },
            },
        ),
        (  # ts21 - mb_67.mpt
            {
                "case": "mb_67.mpt",
                "encoding": "windows-1252",
                "locale": "de_DE.UTF-8",
                "parameters": {"filetype": "eclab.mpt"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 33,
                "point": 0,
                "pars": {
                    "uts": {"value": 1670510213.355},
                    "Ewe": {"value": 2.3278546, "unit": "V", "sigma": 7.5e-05},
                },
            },
        ),
    ],
)
def test_electrochem_from_eclab(input, ts, datadir):
    ret = datagram_from_input(input, "electrochem", datadir, version="5.0")
    standard_datagram_test(ret, ts)
    pars_datagram_test(ret, ts)


@pytest.mark.parametrize(
    "input",
    [
        {  # ts0 - wait.mpr
            "case": "wait.mpr",
            "encoding": "windows-1252",
            "parameters": {"filetype": "eclab.mpr"},
        },
        {  # ts1 - wait.mpt
            "case": "wait.mpt",
            "encoding": "windows-1252",
            "parameters": {"filetype": "eclab.mpt"},
        },
    ],
)
def test_electrochem_wait_technique(input, datadir):
    dg = datagram_from_input(input, "electrochem", datadir)
    assert not dg["0"].data_vars, "No data should be present."


@pytest.mark.parametrize(
    "input, refpath",
    [
        (
            {  # ts0 - ca.mpr
                "case": "ca.mpr",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpr"},
            },
            "ca.dg.json",
        ),
        (
            {  # ts1 - ca.mpt
                "case": "ca.mpt",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpt"},
            },
            "ca.dg.json",
        ),
        (
            {  # ts2 - geis.mpr
                "case": "geis.mpr",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpr"},
            },
            "geis.dg.json",
        ),
        (
            {  # ts3 - geis.mpt
                "case": "geis.mpt",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpt"},
            },
            "geis.dg.json",
        ),
    ],
)
def test_electrochem_compare_raw_values(input, refpath, datadir):
    os.chdir(datadir)
    dg = datagram_from_input(input, "electrochem", datadir)
    step = dg["0"]
    with open(refpath, "r") as infile:
        ref = json.load(infile)["steps"][0]["data"]
    for k in step:
        if k == "uts":
            uts = [ts["uts"] for ts in ref]
            np.testing.assert_allclose(step["uts"], uts)
        elif k.endswith("_std_err"):
            continue
        else:
            res = dg_get_quantity(dg, 0, k)
            if isinstance(res, dict):
                n = [ts["raw"][k]["n"] for ts in ref]
                s = [ts["raw"][k]["s"] for ts in ref]
                u = ref[0]["raw"][k]["u"]
                compare_result_dicts(res, {"n": n, "s": s, "u": u})
            else:
                tv = [ts["raw"][k] for ts in ref]
                assert (res == tv).all()


@pytest.mark.parametrize(
    "input, ts",
    [
        (  # ts0 - vsp_ocv without external devices
            {
                "case": "vsp_ocv_wo.mpr",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpr"},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 121,
                "point": 0,
                "pars": {
                    "Ewe": {
                        "value": 0.03291448,
                        "sigma": 0.00015,
                        "unit": "V",
                    },
                    "uts": {"value": 1641938088.5444725},
                },
            },
        ),
        (  # ts1 - vsp_ocv with external devices
            {
                "case": "vsp_ocv_with.mpr",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpr"},
                "externaldate": {"from": {"utsoffset": 0.0}},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 13,
                "point": 12,
                "pars": {
                    "Ewe": {
                        "value": 0.000106005,
                        "sigma": 0.00015,
                        "unit": "V",
                    },
                    "Force": {
                        "value": 162.9350891,
                        "sigma": 0.0165,
                        "unit": "N",
                    },
                    "Temperatur": {
                        "value": 27.77219582,
                        "sigma": 0.00225,
                        "unit": "°C",
                    },
                    "uts": {"value": 53996.920794260455},
                },
            },
        ),
        (  # ts2 - vsp_peis with external devices
            {
                "case": "vsp_peis_with.mpr",
                "encoding": "windows-1252",
                "parameters": {"filetype": "eclab.mpr"},
                "externaldate": {"from": {"utsoffset": 0.0}},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 2240,
                "point": 840,
                "pars": {
                    "uts": {"value": 527655.5950297173},
                },
            },
        ),
    ],
)
def test_electrochem_vsp_3e(input, ts, datadir):
    ret = datagram_from_input(input, "electrochem", datadir)
    standard_datagram_test(ret, ts)
    pars_datagram_test(ret, ts, atol=1e-8)


@pytest.mark.parametrize(
    "infile, ts",
    [
        (
            "ts0_electrochem_tomato.yml",
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 51,
                "point": 9,
                "pars": {
                    "uts": {"value": 1650490355.3152814},
                    "Ewe": {"value": 4.2000856399, "sigma": 0.00015, "unit": "V"},
                    "I": {"value": 0.0044998624, "sigma": 4e-7, "unit": "A"},
                },
            },
        ),
    ],
)
def test_electrochem_tomato(infile, ts, datadir):
    ret = datagram_from_file(infile, datadir)
    standard_datagram_test(ret, ts)
    pars_datagram_test(ret, ts)


@pytest.mark.parametrize(
    "infile, outfile",
    [
        ("issue_61_Irange_65.mpr", "issue_61.nc"),
        ("issue_61_Irange_65.mpt", "issue_61.nc"),
    ],
)
def test_electrochem_bugs(infile, outfile, datadir):
    os.chdir(datadir)
    if infile.endswith("mpr"):
        filetype = "biologic-mpr"
    elif infile.endswith("mpt"):
        filetype = "biologic-mpt"
    else:
        assert False, "unknown filetype"
    ret = yadg.extractors.extract(filetype=filetype, path=Path(infile))
    ref = xarray.open_dataset(outfile, engine="h5netcdf")
    # ret.to_netcdf(outfile, engine="h5netcdf")
    assert ret["uts"].equals(ref["uts"])
    for k in ret.data_vars:
        if k.endswith("_std_err"):
            continue
        elif ret[k].dtype.kind in {"O", "S", "U"}:
            assert ret[k].equals(ref[k])
        else:
            np.testing.assert_allclose(ret[k], ref[k], equal_nan=True)



import pytest
from tests.utils import (
    datagram_from_input,
    standard_datagram_test,
    pars_datagram_test,
)
import numpy as np


@pytest.mark.parametrize(
    "input, ts",
    [
        (
            {  # ts0 - units on 2nd line, correct number of rows, correct value, sigma from d.p.
                "case": "case_uts_units.csv"
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 6,
                "point": 0,
                "pars": {"flow": {"sigma": 0.1, "value": 15.0, "unit": "ml/min"}},
            },
        ),
        (
            {  # ts1 - timestamp from uts and index
                "case": "case_uts_units.csv",
                "parameters": {"timestamp": {"uts": {"index": 0}}},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 6,
                "point": 2,
                "pars": {"uts": {"value": 1631626610.0}},
            },
        ),
        (
            {  # ts2 - semicolon separator, sigmas from decimal places
                "case": "case_timestamp.ssv",
                "parameters": {
                    "sep": ";",
                    "units": {"flow": "ml/min", "T": "K", "p": "atm"},
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 3,
                "point": 0,
                "pars": {
                    "flow": {"sigma": 0.1, "value": 15.0, "unit": "ml/min"},
                    "T": {"sigma": 0.1, "value": 23.1, "unit": "K"},
                    "p": {"sigma": 0.0001, "value": 1.0001, "unit": "atm"},
                },
            },
        ),
        (
            {  # ts3 - semicolon separator, timestamp from timestamp and index
                "case": "case_timestamp.ssv",
                "parameters": {
                    "sep": ";",
                    "units": {"flow": "ml/min", "T": "K", "p": "atm"},
                    "timestamp": {"timestamp": {"index": 0}},
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 3,
                "point": 1,
                "pars": {"uts": {"value": 1631284345.0}},
            },
        ),
        (
            {  # ts4 - tab separator
                "case": "case_custom_ts.tsv",
                "parameters": {
                    "sep": "\t",
                    "timestamp": {
                        "timestamp": {"index": 1, "format": "%d.%m.%Y %I:%M:%S%p"}
                    },
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 5,
                "point": 3,
                "pars": {"T": {"value": 351.2, "sigma": 0.1, "unit": "K"}},
            },
        ),
        (
            {  # ts5 - tab separator, timestamp from timestamp, format, index
                "case": "case_custom_ts.tsv",
                "parameters": {
                    "sep": "\t",
                    "timestamp": {
                        "timestamp": {"index": 1, "format": "%d.%m.%Y %I:%M:%S%p"}
                    },
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 5,
                "point": 4,
                "pars": {"uts": {"value": 1631280585.0}},
            },
        ),
        (
            {  # ts6 - tab separator, timestamp from timestamp, format, index
                "case": "case_custom_ts.tsv",
                "parameters": {
                    "sep": "\t",
                    "timestamp": {
                        "timestamp": {"index": 1, "format": "%d.%m.%Y %I:%M:%S%p"}
                    },
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 5,
                "point": 2,
                "pars": {"uts": {"value": 1631276985.0}},
            },
        ),
        (
            {  # ts7 - tab separator, timestamp from timestamp, format, index
                "case": "case_custom_ts.tsv",
                "parameters": {
                    "sep": "\t",
                    "timestamp": {
                        "timestamp": {"index": 1, "format": "%d.%m.%Y %I:%M:%S%p"}
                    },
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 5,
                "point": 0,
                "pars": {"uts": {"value": 1631273385.0}},
            },
        ),
        (
            {  # ts8 - tab separator, timestamp from iso date and iso time
                "case": "case_date_time_iso.csv",
                "parameters": {
                    "timestamp": {"date": {"index": 0}, "time": {"index": 1}}
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 4,
                "point": 1,
                "pars": {"uts": {"value": 1610659800.0}},
            },
        ),
        (
            {  # ts9 - timestamp from time with external date only
                "case": "case_time_custom.csv",
                "parameters": {
                    "timestamp": {"time": {"index": 0, "format": "%I.%M%p"}}
                },
                "externaldate": {"from": {"isostring": "2021-01-01"}},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 3,
                "point": 0,
                "pars": {"uts": {"value": 1609502340.0}},
            },
        ),
        (
            {  # ts10 - timestamp from time with custom format only
                "case": "case_time_custom.csv",
                "parameters": {
                    "timestamp": {"time": {"index": 0, "format": "%I.%M%p"}}
                },
                "externaldate": {"from": {"utsoffset": 0.0}},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 3,
                "point": 1,
                "pars": {"uts": {"value": 43200}},
            },
        ),
        (
            {  # ts11 - test stripping
                "case": "picolog_temperature.csv",
                "version": "4.2",
                "parameters": {
                    "sep": ",",
                    "timestamp": {"uts": {"index": 0}},
                    "strip": '"',
                    "units": {
                        "Outside Last (C)": "degC",
                        "Outside Ave. (C)": "degC",
                        "Outside Min. (C)": "degC",
                        "Outside Max. (C)": "degC",
                        "Inside_GDE_TCK01 Last (C)": "degC",
                        "Inside_GDE_TCK01 Ave. (C)": "degC",
                        "Inside_GDE_TCK01 Min. (C)": "degC",
                        "Inside_GDE_TCK01 Max. (C)": "degC",
                        "Inside_Nafion_TCK02 Last (C)": "degC",
                        "Inside_Nafion_TCK02 Ave. (C)": "degC",
                        "Inside_Nafion_TCK02 Min. (C)": "degC",
                        "Inside_Nafion_TCK02 Max. (C)": "degC",
                    },
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 2470,
                "point": 2469,
                "pars": {
                    "Outside Last (C)": {
                        "sigma": 0.01,
                        "value": 26.14,
                        "unit": "degC",
                    }
                },
            },
        ),
        (
            {  # ts12 - test sparse tables
                "case": "picolog_temperature_sparse.csv",
                "version": "4.2",
                "parameters": {
                    "sep": ",",
                    "timestamp": {"uts": {"index": 0}},
                    "strip": '"',
                    "units": {
                        "Outside Last (C)": "degC",
                        "Outside Ave. (C)": "degC",
                        "Outside Min. (C)": "degC",
                        "Outside Max. (C)": "degC",
                        "Inside_GDE_TCK01 Last (C)": "degC",
                        "Inside_GDE_TCK01 Ave. (C)": "degC",
                        "Inside_GDE_TCK01 Min. (C)": "degC",
                        "Inside_GDE_TCK01 Max. (C)": "degC",
                        "Inside_Nafion_TCK02 Last (C)": "degC",
                        "Inside_Nafion_TCK02 Ave. (C)": "degC",
                        "Inside_Nafion_TCK02 Min. (C)": "degC",
                        "Inside_Nafion_TCK02 Max. (C)": "degC",
                    },
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 2226,
                "point": 361,
                "pars": {
                    "Outside Last (C)": {
                        "sigma": 0.01,
                        "value": 25.97,
                        "unit": "degC",
                    },
                    "Inside_Nafion_TCK02 Max. (C)": {
                        "value": np.nan,
                        "sigma": np.nan,
                        "unit": "degC",
                    },
                },
            },
        ),
    ],
)
def test_datagram_from_basiccsv(input, ts, datadir):
    ver = input.pop("version", "4.0")
    ret = datagram_from_input(input, "basiccsv", datadir, version=ver)
    standard_datagram_test(ret, ts)
    pars_datagram_test(ret, ts)


def test_issue_143(datadir):
    input = {
        "case": "flow_data.csv",
        "version": "4.2",
        "parameters": {
            "sep": ",",
            "timestamp": {"time": {"index": 5}},
            "units": {
                "DryCal smL/min": "smL/min",
                "DryCal Avg. smL/min": "smL/min",
                "Temp. Deg C": "degC",
                "Pressure mBar": "mbar",
            },
        },
    }
    ver = input.pop("version", "4.0")
    ret = datagram_from_input(input, "basiccsv", datadir, version=ver)
    print(f"{ret=}")
    assert ret["0"]["DryCal smL_min"].attrs["units"] == "smL/min"



"""
Handles the reading and processing of mass spectrometry files. The basic function of the
parser is to:

#. read in the raw data and create timestamped traces with one :class:`xarray.Dataset` per trace
#. collect `metadata` such as the software version, author, etc.

Usage
`````
Select :mod:`~yadg.parsers.masstrace` by supplying it to the ``parser`` keyword
in the `dataschema`. The parser supports the following parameters:

.. _yadg.parsers.masstrace.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_0.step.MassTrace

.. _yadg.parsers.masstrace.formats:

Formats
```````
The ``filetypes`` currently supported by the parser are:

 - Pfeiffer Quadstar 32-bit scan analog data (``quadstar.sac``),
   see :mod:`~yadg.parsers.masstrace.quadstarsac`

.. _yadg.parsers.masstrace.provides:

Schema
``````
The raw data, loaded from the supplied files, is stored using the following format:

.. code-block:: yaml

  datatree.DataTree:
    {{ detector_name }}   !!xr.Dataset
      coords:
        uts:              !!float
        mass_to_charge:   !!float                 # m/z (amu)
      data_vars:
        y:                (uts, mass_to_charge)   # Detected signal (counts)

The uncertainties in ``mass_to_charge`` are taken as the step-width of
the linearly spaced mass values.

The uncertainties in of ``y`` are the largest value between:

#. The quantization error from the ADC, its resolution assumed to be 32
   bit. Dividing F.S.R. by ``2 ** 32`` gives an error in the order of
   magnitude of the smallest data value in ``y``.
#. The contribution from neighboring masses. In the operating manual of
   the QMS 200 (see 2.8 QMS 200 F & 2.9 QMS 200 M), a maximum
   contribution from the neighboring mass of 50 ppm is noted.

.. note::

    The data in ``y`` may contain ``NaN`` s. The measured ion
    count/current value will occasionally exceed the specified detector
    F.S.R. (e.g. 1e-9), and will then flip directly to the maximum value
    of a float32. These values are set to ``float("NaN")``.

Module Functions
````````````````

"""

import datatree
from . import quadstarsac


def process(
    *,
    filetype: str,
    **kwargs: dict,
) -> datatree.DataTree:
    """
    Unified mass spectrometry data parser.Forwards ``kwargs`` to the worker functions
    based on the supplied ``filetype``.

    Parameters
    ----------
    filetype
        Discriminator used to select the appropriate worker function.

    Returns
    -------
    :class:`datatree.DataTree`

    """
    if filetype == "quadstar.sac":
        return quadstarsac.process(**kwargs)



import logging
from uncertainties.core import str_to_number_with_uncert as tuple_fromstr
from typing import Callable, Any
from pydantic import BaseModel
import locale as lc
from ... import dgutils

import numpy as np
import xarray as xr

logger = logging.getLogger(__name__)


def process_row(
    headers: list,
    items: list,
    datefunc: Callable,
    datecolumns: list,
) -> tuple[dict, dict]:
    """
    A function that processes a row of a table.

    This is the main worker function of :mod:`~yadg.parsers.basiccsv`, but is often
    re-used by any other parser that needs to process tabular data.

    Parameters
    ----------
    headers
        A list of headers of the table.

    items
        A list of values corresponding to the headers. Must be the same length as headers.

    units
        A dict for looking up the units corresponding to a certain header.

    datefunc
        A function that will generate ``uts`` given a list of values.

    datecolumns
        Column indices that need to be passed to ``datefunc`` to generate uts.

    Returns
    -------
    vals, devs
        A tuple of result dictionaries, with the first element containing the values
        and the second element containing the deviations of the values.

    """
    assert len(headers) == len(items), (
        f"process_row: Length mismatch between provided headers: "
        f"{headers} and  provided items: {items}."
    )

    vals = {}
    devs = {}
    columns = [column.strip() for column in items]

    # Process raw data, assign sigma and units
    vals["uts"] = datefunc(*[columns[i] for i in datecolumns])
    for ci, header in enumerate(headers):
        if ci in datecolumns:
            continue
        elif columns[ci] == "":
            continue
        try:
            val, dev = tuple_fromstr(lc.delocalize(columns[ci]))
            vals[header] = val
            devs[header] = dev
        except ValueError:
            vals[header] = columns[ci]

    return vals, devs


def append_dicts(
    vals: dict[str, Any],
    devs: dict[str, Any],
    data: dict[str, list[Any]],
    meta: dict[str, list[Any]],
    fn: str = None,
    li: int = 0,
) -> None:
    if "_fn" in meta and fn is not None:
        meta["_fn"].append(str(fn))
    for k, v in vals.items():
        if k not in data:
            data[k] = [None if isinstance(v, str) else np.nan] * li
        data[k].append(v)
    for k, v in devs.items():
        if k not in meta:
            meta[k] = [np.nan] * li
        meta[k].append(v)

    for k in set(data) - set(vals):
        data[k].append(np.nan)
    for k in set(meta) - set(devs):
        if k != "_fn":
            meta[k].append(np.nan)


def dicts_to_dataset(
    data: dict[str, list[Any]],
    meta: dict[str, list[Any]],
    units: dict[str, str] = dict(),
    fulldate: bool = True,
) -> xr.Dataset:
    darrs = {}
    for key, val in data.items():
        attrs = {}
        u = units.get(key, None)
        if u is not None:
            attrs["units"] = u
        if key == "uts":
            continue
        if "/" in key:
            logger.warning(f"Replacing '/' for '_' in column {key!r}.")
            k = key.replace("/", "_")
        else:
            k = key
        darrs[k] = xr.DataArray(data=val, dims=["uts"], attrs=attrs)
        if key in meta and darrs[k].dtype.kind in {"i", "u", "f", "c", "m", "M"}:
            err = f"{k}_std_err"
            darrs[k].attrs["ancillary_variables"] = err
            attrs["standard_name"] = f"{k} standard error"
            darrs[err] = xr.DataArray(data=meta[key], dims=["uts"], attrs=attrs)
    if "uts" in data:
        coords = dict(uts=data.pop("uts"))
    else:
        coords = dict()
    if fulldate:
        attrs = dict()
    else:
        attrs = dict(fulldate=False)
    return xr.Dataset(data_vars=darrs, coords=coords, attrs=attrs)


def process(
    *,
    fn: str,
    encoding: str,
    locale: str,
    timezone: str,
    parameters: BaseModel,
    **kwargs: dict,
) -> xr.Dataset:
    """
    A basic csv parser.

    This parser processes a csv file. The header of the csv file consists of one or two
    lines, with the column headers in the first line and the units in the second. The
    parser also attempts to parse column names to produce a timestamp, and save all other
    columns as floats or strings.

    Parameters
    ----------
    fn
        File to process

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "localtime".

    parameters
        Parameters for :class:`~dgbowl_schemas.yadg.dataschema_5_0.step.BasicCSV`.

    Returns
    -------
    :class:`xarray.Dataset`
        No metadata is returned by the :mod:`~yadg.parsers.basiccsv` parser. The full
        date might not be returned, eg. when only time is specified in columns.

    """

    if hasattr(parameters, "strip"):
        strip = parameters.strip
    else:
        strip = None

    # Load file, extract headers and get timestamping function
    with open(fn, "r", encoding=encoding) as infile:
        # This decode/encode is done to account for some csv files that have a BOM
        # at the beginning of each line.
        lines = [i.encode().decode(encoding) for i in infile.readlines()]
    assert len(lines) >= 2
    headers = [h.strip().strip(strip) for h in lines[0].split(parameters.sep)]

    datecolumns, datefunc, fulldate = dgutils.infer_timestamp_from(
        headers=headers, spec=parameters.timestamp, timezone=timezone
    )

    # Populate units
    units = parameters.units
    if units is None:
        units = {}
        _units = [c.strip().strip(strip) for c in lines[1].split(parameters.sep)]
        for header in headers:
            units[header] = _units.pop(0)
        si = 2
    else:
        for header in headers:
            if header not in units:
                logger.warning(
                    "Using implicit dimensionless unit ' ' for '%s'.", header
                )
                units[header] = " "
            elif units[header] == "":
                units[header] = " "
        si = 1

    units = dgutils.sanitize_units(units)

    # Process rows
    old_loc = lc.getlocale(category=lc.LC_NUMERIC)
    lc.setlocale(lc.LC_NUMERIC, locale=locale)
    data_vals = {}
    meta_vals = {"_fn": []}
    for li, line in enumerate(lines[si:]):
        vals, devs = process_row(
            headers,
            [i.strip().strip(strip) for i in line.split(parameters.sep)],
            datefunc,
            datecolumns,
        )
        append_dicts(vals, devs, data_vals, meta_vals, fn, li)
    lc.setlocale(category=lc.LC_NUMERIC, locale=old_loc)

    return dicts_to_dataset(data_vals, meta_vals, units, fulldate)



import pytest
import os
import yaml
from tests.utils import (
    datagram_from_input,
    standard_datagram_test,
    compare_result_dicts,
    dg_get_quantity,
)
from yadg.core import process_schema
from dgbowl_schemas.yadg import to_dataschema


def special_datagram_test(datagram, testspec):
    for i in {"height", "area", "concentration", "xout", "retention time"}:
        if i not in testspec:
            continue
        ret = dg_get_quantity(datagram, testspec["step"], i, testspec["point"])
        for k, v in testspec[i].items():
            rval = ret["n"].sel(dict(species=k))
            rdev = ret["s"].sel(dict(species=k))
            ru = ret["u"]
            print(f"{testspec[i]=}")
            print(f"{ret['n'].sel(dict(species=k))=}")
            compare_result_dicts({"n": rval, "s": rdev, "u": ru}, v)
        assert ret["n"].uts == testspec["uts"]


@pytest.mark.parametrize(
    "input, ts",
    [
        (
            {  # ts0 - fusion-json
                "folders": ["."],
                "prefix": "15p",
                "suffix": "fusion-data",
                "parameters": {
                    "filetype": "fusion.json",
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 3,
                "method": "AS_Cal_20220204",
                "point": 1,
                "height": {
                    "MeOH": {"n": 0.0, "s": 1.0},
                },
                "area": {
                    "O2": {"n": 1034.53, "s": 0.01, "u": None},
                },
                "concentration": {
                    "CO": {"n": 0.02911539, "s": 0.0000291, "u": "%"},
                },
                "xout": {
                    "C2H6": {"n": 0.22425762, "s": 0.00022426, "u": "%"},
                },
                "retention time": {
                    "C2H6": {"n": 43.78, "s": 0.01, "u": "s"},
                },
                "uts": 1654697826.467,
            },
        ),
        (
            {  # ts1 - fusion-zip
                "folders": ["."],
                "prefix": "20220608",
                "suffix": "zip",
                "parameters": {
                    "filetype": "fusion.zip",
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 15,
                "method": "AS_Cal_20220204",
                "point": 13,
                "height": {
                    "MeOH": {"n": 0.0, "s": 1.0},
                },
                "area": {
                    "O2": {"n": 1034.53, "s": 0.01},
                },
                "concentration": {
                    "CO": {"n": 0.02911539, "s": 0.0000291, "u": "%"},
                },
                "xout": {
                    "C2H6": {"n": 0.22425762, "s": 0.00022426, "u": "%"},
                },
                "retention time": {
                    "C2H6": {"n": 43.78, "s": 0.01, "u": "s"},
                },
                "uts": 1654697826.467,
            },
        ),
        (
            {  # ts1 - fusion-csv
                "folders": ["."],
                "prefix": "20220608",
                "suffix": "csv",
                "parameters": {
                    "filetype": "fusion.csv",
                },
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 15,
                "method": "AS_Cal_20220204",
                "point": 1,
                "area": {
                    "O2": {"n": 1034.53, "s": 0.001, "u": None},
                },
                "concentration": {
                    "CO": {"n": 0.02911539, "s": 1e-8, "u": "%"},
                },
                "xout": {
                    "C2H6": {"n": 0.22425762, "s": 1e-8, "u": "%"},
                },
                "retention time": {
                    "C2H6": {"n": 43.78, "s": 0.01, "u": "s"},
                },
                "uts": 1654697826.0,
            },
        ),
        (
            {  # ts2 - empalc.csv
                "folders": ["."],
                "suffix": "20p_v2.csv",
                "parameters": {
                    "filetype": "empalc.csv",
                },
                "externaldate": {"using": {"utsoffset": 0}},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 18,
                "method": "CO2RR_ChA_FTI_0.6mL_40uL_45dgr_32min.amx",
                "point": 17,
                "height": {
                    "Formic acid": {"n": 3436.903, "s": 0.001, "u": None},
                },
                "area": {
                    "Formic acid": {"n": 155913.098, "s": 0.001, "u": None},
                },
                "concentration": {
                    "Formic acid": {"n": 18.7521, "s": 0.0001, "u": "mmol/l"},
                },
                "retention time": {
                    "Formic acid": {"n": 15.671, "s": 0.001, "u": "min"},
                },
                "uts": 20400.0,
            },
        ),
        (
            {  # ts2 - empalc.csv
                "folders": ["."],
                "suffix": "25p_v2.csv",
                "parameters": {
                    "filetype": "empalc.csv",
                },
                "externaldate": {"using": {"utsoffset": 0}},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 19,
                "method": "CO2RR_ChA_FTI_0.6mL_40uL_45dgr_32min.amx",
                "point": 18,
                "height": {
                    "Ethanol": {"n": 25495.948, "s": 0.001, "u": None},
                },
                "area": {
                    "Ethanol": {"n": 2745541.097, "s": 0.001, "u": None},
                },
                "concentration": {
                    "Ethanol": {"n": 254.4736, "s": 0.0001, "u": "mmol/l"},
                },
                "retention time": {
                    "Ethanol": {"n": 21.098, "s": 0.001, "u": "min"},
                },
                "uts": 21600.0,
            },
        ),
        (
            {  # ts2 - empalc.xlsx
                "folders": ["."],
                "suffix": "25p_v2.xlsx",
                "parameters": {
                    "filetype": "empalc.xlsx",
                },
                "externaldate": {"using": {"utsoffset": 0}},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 19,
                "method": "CO2RR_ChA_FTI_0.6mL_40uL_45dgr_32min.amx",
                "point": 18,
                "height": {
                    "Ethanol": {"n": 25495.948, "s": 0.001},
                },
                "area": {
                    "Ethanol": {"n": 2745541.097, "s": 0.001, "u": None},
                },
                "concentration": {
                    "Ethanol": {"n": 254.4736, "s": 0.0001, "u": "mmol/l"},
                },
                "retention time": {
                    "Ethanol": {"n": 21.098, "s": 0.001, "u": "min"},
                },
                "uts": 21600.0,
            },
        ),
        (
            {  # ts3 - empalc.xlsx with newlines
                "folders": ["."],
                "suffix": "newlines.xlsx",
                "parameters": {
                    "filetype": "empalc.xlsx",
                },
                "externaldate": {"using": {"utsoffset": 0}},
            },
            {
                "nsteps": 1,
                "step": 0,
                "nrows": 20,
                "method": "CO2RR_ChA_FTI_0.6mL_40uL_45dgr_32min.amx",
                "point": 18,
                "height": {
                    "Ethanol": {"n": 464.048, "s": 0.001},
                },
                "area": {
                    "Ethanol": {"n": 26514.098, "s": 0.001},
                },
                "concentration": {
                    "Ethanol": {"n": 2.4303, "s": 0.0001, "u": "mmol/l"},
                },
                "retention time": {
                    "Ethanol": {"n": 20.74, "s": 0.01, "u": "min"},
                },
                "uts": 21600.0,
            },
        ),
    ],
)
def test_datagram_from_chromdata(input, ts, datadir):
    os.chdir(datadir)
    ret = datagram_from_input(input, "chromdata", datadir, version="4.2")
    standard_datagram_test(ret, ts)
    special_datagram_test(ret, ts)


def test_lock_stock_chromdata(datadir):
    os.chdir(datadir)
    with open("lock_stock_dataschema.yml", "r") as inf:
        schema = yaml.safe_load(inf)
    ret = process_schema(to_dataschema(**schema))
    print(f"{ret=}")
    for k in {"height", "concentration", "retention time", "area"}:
        assert ret["LC"][k].shape == (7, 2)



"""
This module handles the reading and processing of files containing electrochemical
data, including BioLogic's EC-Lab file formats. The basic function of the parser is to:

#. Read in the technique data and create timesteps.
#. Collect metadata, such as the measurement settings and the loops
   contained in a given file.
#. Collect data describing the technique parameter sequences.

Usage
`````
Available since ``yadg-4.0``. The parser supports the following parameters:

.. _yadg.parsers.electrochem.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_0.step.ElectroChem

.. _yadg.parsers.electrochem.formats:

Formats
```````
The ``filetypes`` currently supported by the parser are:

 - EC-Lab raw data binary file and parameter settings (``eclab.mpr``),
   see :mod:`~yadg.parsers.electrochem.eclabmpr`
 - EC-Lab human-readable text export of data (``eclab.mpt``),
   see :mod:`~yadg.parsers.electrochem.eclabmpt`
 - tomato's structured json output (``tomato.json``),
   see :mod:`~yadg.parsers.electrochem.tomatojson`

Schema
``````
Depending on the filetype, the output :class:`xarray.Dataset` may contain multiple
derived values. However, all filetypes will report at least the following:

.. code-block:: yaml

  xr.Dataset:
    coords:
      uts:      !!float
    data_vars:
      Ewe:      (uts)    # Potential of the working electrode (V)
      Ece:      (uts)    # Potential of the counter electrode (V)
      I:        (uts)    # Applied current (A)

In some cases, average values (i.e. ``<Ewe>`` or ``<I>``) may be reported instead
of the instantaneous data.

.. warning::

  In previous versions of :mod:`yadg`, the :mod:`~yadg.parsers.electrochem` parser
  optionally transposed data from impedance spectroscopy, grouping the datapoints
  in each scan into a single "trace". This behaviour has been removed in ``yadg-5.0``.

Module Functions
````````````````

"""

import xarray as xr
from . import eclabmpr, eclabmpt, tomatojson


def process(
    *,
    filetype: str,
    **kwargs: dict,
) -> xr.Dataset:
    """
    Unified parser for electrochemistry data. Forwards ``kwargs`` to the worker functions
    based on the supplied ``filetype``.

    Parameters
    ----------
    filetype
        Discriminator used to select the appropriate worker function.

    Returns
    -------
    :class:`xarray.Dataset`

    """
    if filetype == "eclab.mpr":
        return eclabmpr.process(**kwargs)
    elif filetype == "eclab.mpt":
        return eclabmpt.process(**kwargs)
    elif filetype == "tomato.json":
        return tomatojson.process(**kwargs)



"""
**fusionjson**: Processing Inficon Fusion json data format (json).
------------------------------------------------------------------

This is a fairly detailed data format, including the traces, the calibration applied,
and also the integrated peak areas and other processed information, which are parsed
by this module.

.. note ::

    To parse the raw trace data, use the :mod:`~yadg.parsers.chromtrace` module.

.. warning ::

    The detectors in the json files are not necessarily in a consistent order. To
    avoid inconsistent parsing of species which appear in both detectors, the
    detector keys are sorted. **Species present in both detectors** will be
    **overwritten by the last detector** in alphabetical order.

Exposed metadata:
`````````````````

.. code-block:: yaml

    params:
      method:   !!str
      username: None
      version:  !!str
      datafile: !!str

.. codeauthor:: Peter Kraus
"""

import json
import logging
from zoneinfo import ZoneInfo
from ...dgutils.dateutils import str_to_uts
import xarray as xr
import numpy as np

logger = logging.getLogger(__name__)


def process(
    *, fn: str, encoding: str, timezone: ZoneInfo, **kwargs: dict
) -> xr.Dataset:
    """
    Fusion json format.

    One chromatogram per file with multiple traces, and pre-analysed results.
    Only a subset of the metadata is retained, including the method name,
    detector names, and information about assigned peaks.

    Parameters
    ----------
    fn
        Filename to process.

    encoding
        Encoding used to open the file.

    timezone
        Timezone information. This should be ``"localtime"``.

    Returns
    -------
    :class:`xarray.Dataset`

    """

    with open(fn, "r", encoding=encoding, errors="ignore") as infile:
        jsdata = json.load(infile)
    metadata = {
        "method": jsdata.get("methodName", "n/a"),
        "version": jsdata.get("softwareVersion", {}).get("version", None),
        "datafile": jsdata.get("sequence", {}).get("location", None),
    }
    uts = str_to_uts(timestamp=jsdata["runTimeStamp"], timezone=timezone)

    sampleid = jsdata.get("annotations", {}).get("name", None)
    if sampleid is not None:
        metadata["sampleid"] = sampleid

    units = {
        "height": None,
        "area": None,
        "concentration": "%",
        "xout": "%",
        "retention time": "s",
    }

    raw = {
        "height": {},
        "area": {},
        "concentration": {},
        "xout": {},
        "retention time": {},
    }

    species = set()

    # sort detector keys to ensure alphabetic order for ID matching
    for detname in sorted(jsdata["detectors"].keys()):
        detdict = jsdata["detectors"][detname]
        if "analysis" in detdict:
            for peak in detdict["analysis"]["peaks"]:
                if "label" not in peak:
                    continue
                else:
                    species.add(peak["label"])
                if "height" in peak:
                    raw["height"][peak["label"]] = (float(peak["height"]), 1.0)
                if "area" in peak:
                    raw["area"][peak["label"]] = (float(peak["area"]), 0.01)
                if "concentration" in peak:
                    raw["concentration"][peak["label"]] = (
                        float(peak["concentration"]),
                        float(peak["concentration"]) * 1e-3,
                    )
                if "normalizedConcentration" in peak:
                    raw["xout"][peak["label"]] = (
                        float(peak["normalizedConcentration"]),
                        float(peak["normalizedConcentration"]) * 1e-3,
                    )
                if "top" in peak:
                    raw["retention time"][peak["label"]] = (float(peak["top"]), 0.01)
        else:
            logger.warning("'analysis' of chromatogram not present in file '%s'", fn)

    valve = jsdata.get("annotations", {}).get("valcoPosition", None)
    if valve is not None:
        raw["valve"] = valve

    species = sorted(species)
    data_vars = {}
    for k, v in units.items():
        vals, devs = zip(*[raw[k].get(s, (np.nan, np.nan)) for s in species])
        data_vars[k] = (
            ["uts", "species"],
            [vals],
            {"ancillary_variables": f"{k}_std_err"},
        )
        data_vars[f"{k}_std_err"] = (
            ["uts", "species"],
            [devs],
            {"standard_name": f"{k} stdandard_error"},
        )
        if v is not None:
            data_vars[k][2]["units"] = v
            data_vars[f"{k}_std_err"][2]["units"] = v

    ds = xr.Dataset(
        data_vars=data_vars,
        coords={"species": (["species"], species), "uts": (["uts"], [uts])},
        attrs=metadata,
    )
    return ds
